from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.registration.model.translation_registration_model import TranslationRegistrationModel
from dexp.processing.synthetic_datasets.multiview_data import generate_fusion_test_data


def test_translation_model_numpy():
    with NumpyBackend():
        _test_translation_model()


def test_translation_model_cupy():
    try:
        with CupyBackend():
            _test_translation_model()
    except ModuleNotFoundError:
        print("Cupy module not found! demo ignored")


def _test_translation_model(length_xy=128):
    xp = Backend.get_xp_module()

    image_gt, image_lowq, blend_a, blend_b, image1, image2 = generate_fusion_test_data(add_noise=False,
                                                                                       shift=(1, 5, -13),
                                                                                       volume_fraction=0.5,
                                                                                       length_xy=length_xy,
                                                                                       length_z_factor=2,
                                                                                       amount_low=0,
                                                                                       zero_level=0)

    model = TranslationRegistrationModel(shift_vector=(-1, -5, 13))

    image1_reg, image2_reg = model.apply_pair(image1, image2, pad=False)
    dumb_fusion = xp.maximum(image1_reg, image2_reg)
    average_error = xp.mean(xp.absolute(dumb_fusion - image_gt))
    print(f"average_error = {average_error}")

    image1_reg_pad, image2_reg_pad = model.apply_pair(image1, image2, pad=True)
    image1_reg_pad = image1_reg_pad[0:length_xy // 2, 0:length_xy, 0:length_xy]
    image2_reg_pad = image2_reg_pad[0:length_xy // 2, 0:length_xy, 0:length_xy]
    dumb_fusion_pad = xp.maximum(image1_reg_pad, image2_reg_pad)
    image_gt_shifted = xp.roll(image_gt, shift=(1, 5, 0), axis=(0, 1, 2))
    image_gt_shifted = image_gt_shifted[0:length_xy // 2, 0:length_xy, 0:length_xy]
    average_error_pad = xp.mean(xp.absolute(dumb_fusion_pad - image_gt_shifted))
    print(f"average_error_pad = {average_error_pad}")

    # from napari import Viewer, gui_qt
    # with gui_qt():
    #     def _c(array):
    #         return backend.to_numpy(array)
    #     viewer = Viewer()
    #     viewer.add_image(_c(image1), name='image1')
    #     viewer.add_image(_c(image2), name='image2')
    #     viewer.add_image(_c(image1_reg), name='image1_reg')
    #     viewer.add_image(_c(image2_reg), name='image2_reg')
    #     viewer.add_image(_c(dumb_fusion), name='dumb_fusion')
    #     viewer.add_image(_c(image_gt), name='image_gt')
    #     viewer.add_image(_c(image1_reg_pad), name='image1_reg_pad')
    #     viewer.add_image(_c(image2_reg_pad), name='image2_reg_pad')
    #     viewer.add_image(_c(dumb_fusion_pad), name='dumb_fusion_pad')
    #     viewer.add_image(_c(image_gt_shifted), name='image_gt_shifted')

    assert average_error < 11
    assert average_error_pad < 11



# You need to point to a tiff file with 4 views as first dim,
# as produced for example by: dexp tiff -w -s [128:129] dataset.zarr -o /home/royer/Desktop/test_data/test_data.tiff
from napari import gui_qt, Viewer
from tifffile import imread

from dexp.optics.psf.standard_psfs import nikon16x08na
from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.deconvolution.lr_deconvolution import lucy_richardson_deconvolution
from dexp.processing.restoration.dehazing import dehaze
from dexp.processing.utils.scatter_gather_i2i import scatter_gather_i2i
from dexp.utils.timeit import timeit

filepath = '/home/royer/Desktop/test_data/embryo_fused.tif'


def demo_simview_deconv_numpy():
    with NumpyBackend():
        simview_deconv()


def demo_simview_deconv_cupy():
    try:
        with CupyBackend(enable_unified_memory=True):
            simview_deconv()
    except ModuleNotFoundError:
        print("Cupy module not found! demo ignored")


def simview_deconv():
    xp = Backend.get_xp_module()

    print(f"Loading data...")
    array = imread(filepath)
    print(f"Done loading.")

    # array = array[150 - 64:150 + 64,
    #         1533 - 256:1533 + 256,
    #         931 - 256:931 + 256]

    view = Backend.to_backend(array, dtype=xp.float32)

    # with timeit(f"Clip view ..."):
    #     view = xp.clip(view, a_min=0, a_max=3048, out=view)

    with timeit(f"Dehaze view ..."):
        view_dehazed = dehaze(view, size=65, minimal_zero_level=0)

    min_value = view_dehazed.min()
    max_value = view_dehazed.max()

    print(f"min={min_value}, max={max_value}")

    psf_size = 31
    psf = nikon16x08na(xy_size=psf_size, z_size=psf_size)

    parameters = {'psf': psf,
                  'normalise_input': True,
                  'normalise_minmax': (min_value, max_value),
                  'max_correction': None,
                  'power': 1.0,
                  # 'blind_spot': 3,
                  # 'blind_spot_mode': 'median+uniform',
                  # 'blind_spot_axis_exclusion': (0,),
                  'clip_output': False}

    parameters_wb = {'back_projection': 'wb',
                     'wb_cutoffs': 0.9,
                     'wb_order': 2,
                     'wb_beta': 0.05, }

    with timeit("lucy_richardson_deconvolution"):
        def f(_image):
            return lucy_richardson_deconvolution(image=_image, num_iterations=10, **parameters)

        view_dehazed_deconvolved = scatter_gather_i2i(f, view_dehazed, chunks=320, margins=psf_size, clip=False, to_numpy=True)

    with timeit("lucy_richardson_deconvolution_wb"):
        def f(_image):
            return lucy_richardson_deconvolution(image=_image, num_iterations=3, **parameters, **parameters_wb)

        view_dehazed_deconvolved_wb = scatter_gather_i2i(f, view_dehazed, chunks=320, margins=psf_size, clip=False, to_numpy=True)

    with gui_qt():
        def _c(array):
            return Backend.to_numpy(array)

        viewer = Viewer()
        viewer.add_image(_c(view), name='raw',
                         contrast_limits=(0, 2000), scale=(4, 1, 1), visible=False)
        viewer.add_image(_c(view_dehazed), name='dehazed',
                         contrast_limits=(0, 2000), scale=(4, 1, 1), colormap='bop purple', blending='additive', visible=False)
        viewer.add_image(_c(view_dehazed_deconvolved), name='view_dehazed_deconvolved',
                         contrast_limits=(0, 13000), scale=(4, 1, 1), colormap='bop blue', blending='additive')
        viewer.add_image(_c(view_dehazed_deconvolved_wb), name='view_dehazed_deconvolved_wb',
                         contrast_limits=(0, 13000), scale=(4, 1, 1), colormap='bop blue', blending='additive')


if __name__ == "__main__":
    # demo_simview_fuse_numpy()
    demo_simview_deconv_cupy()



from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.color.demo.demo_scale_bar import demo_scale_bar


def test_scale_bar_numpy():
    with NumpyBackend():
        _test_scale_bar()


def test_scale_bar_cupy():
    try:
        with CupyBackend():
            _test_scale_bar()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _test_scale_bar():
    demo_scale_bar(display=False)



import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.synthetic_datasets.nuclei_background_data import generate_nuclei_background_data
from dexp.processing.utils.normalise import normalise_functions


def test_normalise_numpy():
    with NumpyBackend():
        _test_normalise()


def test_normalise_cupy():
    try:
        with CupyBackend():
            _test_normalise()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _test_normalise(length_xy=128):
    xp = Backend.get_xp_module()

    _, _, image = generate_nuclei_background_data(add_noise=True,
                                                  length_xy=length_xy,
                                                  length_z_factor=4,
                                                  dtype=numpy.float32)

    image = image.astype(numpy.uint16)

    norm_fun, denorm_fun = normalise_functions(image, low=-0.5, high=1, in_place=False, clip=True, dtype=numpy.float32)

    image_normalised = norm_fun(image)
    image_denormalised = denorm_fun(image_normalised)

    assert image_normalised.dtype == numpy.float32
    assert image_denormalised.dtype == image.dtype

    assert image_normalised.shape == image.shape
    assert image_denormalised.shape == image.shape

    assert image_normalised.min() >= -0.5
    assert image_normalised.max() <= 1
    assert image_normalised.max() - image_normalised.min() >= 1.5

    assert image_denormalised.min() * (1 + 1e-3) >= image.min()
    assert image_denormalised.max() <= (1 + 1e-3) * image.max()
    assert (image_denormalised.max() - image_denormalised.min()) * (1 + 1e-3) >= image.max() - image.min()

    image = Backend.to_numpy(image)
    image_denormalised = Backend.to_numpy(image_denormalised)
    error = numpy.median(numpy.abs(image - image_denormalised))
    print(f"error={error}")
    assert error < 0.02

    # from napari import Viewer, gui_qt
    # with gui_qt():
    #     viewer = Viewer()
    #     viewer.add_image(image_gt, name='image_gt')
    #     viewer.add_image(image1, name='image1')
    #     viewer.add_image(image2, name='image2')
    #     viewer.add_image(blend_a, name='blend_a')
    #     viewer.add_image(blended, name='blended')



from pprint import pprint

import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.interpolation.warp import warp
from dexp.utils.timeit import timeit


def demo_warp_1d_numpy():
    try:
        with NumpyBackend():
            _demo_warp_1d()
    except NotImplementedError:
        print("Numpy version not yet implemented")


def demo_warp_1d_cupy():
    try:
        with CupyBackend():
            _demo_warp_1d()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _demo_warp_1d():
    image = numpy.random.uniform(low=0, high=1, size=(128,)).astype(dtype=numpy.float32)

    magnitude = 15
    vector_field = numpy.random.uniform(low=-magnitude, high=+magnitude, size=(8,))

    with timeit("warp"):
        warped = warp(image, vector_field)

    with timeit("dewarp"):
        dewarped = warp(warped, -vector_field)

    warped = Backend.to_numpy(warped)
    dewarped = Backend.to_numpy(dewarped)

    pprint(warped - image)
    pprint(dewarped - image)

    pprint(numpy.max(numpy.absolute(warped - image)))
    pprint(numpy.max(numpy.absolute(dewarped - image)))


if __name__ == "__main__":
    demo_warp_1d_cupy()
    # demo_warp_1d_numpy()



from typing import Union, Tuple

from arbol import section, aprint

from dexp.processing.backends.backend import Backend
from dexp.processing.registration.model.warp_registration_model import WarpRegistrationModel
from dexp.processing.registration.translation_nd_proj import register_translation_proj_nd
from dexp.processing.utils.scatter_gather_i2v import scatter_gather_i2v


@section("register_warp_nd")
def register_warp_nd(image_a,
                     image_b,
                     chunks: Union[int, Tuple[int, ...]],
                     margins: Union[int, Tuple[int, ...]] = None,
                     registration_method=register_translation_proj_nd,
                     force_numpy: bool = False,
                     **kwargs) -> WarpRegistrationModel:
    """
    Registers two nD images using warp model (piece-wise translation model).


    Parameters
    ----------
    image_a : First image to register
    image_b : Second image to register
    chunks : Chunk sizes to divide image into
    margins : Margins to add along each dimension per chunk
    registration_method : registration method to use per tile, must return a TranslationRegistrationModel.
    force_numpy: Forces output model to be allocated with numpy arrays.
    all additional kwargs are passed to the registration method (by default register_translation_maxproj_nd)

    Returns
    -------
    Warp registration model

    """
    image_a = Backend.to_backend(image_a)
    image_b = Backend.to_backend(image_b)

    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    def f(x, y):
        model = registration_method(x, y, force_numpy=force_numpy, **kwargs)
        aprint(f"model: {model} {'' if model.confidence > 0.3 else '(LOW QUALITY!)'}")
        shift, confidence = model.get_shift_and_confidence()
        return xp.asarray(shift), xp.asarray(confidence)

    vector_field, confidence = scatter_gather_i2v(f,
                                                  (image_a, image_b),
                                                  chunks=chunks,
                                                  margins=margins)

    model = WarpRegistrationModel(vector_field=vector_field, confidence=confidence, force_numpy=force_numpy)

    return model



import numpy
from arbol import aprint, asection
from numpy import random

from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.utils.linear_solver import linsolve


def test_linear_solver_numpy():
    with NumpyBackend():
        _test_linear_solver_small()
        # _test_linear_solver_large()
        # test_linear_solver_compare()


def test_linear_solver_cupy():
    try:
        with CupyBackend():
            _test_linear_solver_small()
            # _test_linear_solver_large()
            # test_linear_solver_compare()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _test_linear_solver_small():
    w = 37
    h = 41
    noise = 1e-2

    a = random.rand(w, h)
    a *= random.rand(w, h) > 0.9
    x_gt = random.rand(h)
    x_gt *= random.rand(h) > 0.5
    y_obs = a @ x_gt + noise * (random.rand(w) - 0.5)

    with asection("Running l2 solver on small matrix: "):
        aprint(f"error = {numpy.mean(numpy.absolute(linsolve(a, y_obs, order_error=2, alpha_reg=0) - x_gt))}")

    # with asection("Running l1 solver on small matrix: "):
    #     aprint(f"error = {numpy.mean(numpy.absolute(linsolve(a, y_obs, order_error=1, alpha_reg=0) - x_gt))}")
    #
    # with asection("Running l0.5 solver on small matrix: "):
    #     aprint(f"error = {numpy.mean(numpy.absolute(linsolve(a, y_obs, order_error=0.5, alpha_reg=0) - x_gt))}")
    #
    # with asection("Running l2l1 solver on small matrix: "):
    #     aprint(f"error = {numpy.mean(numpy.absolute(linsolve(a, y_obs, order_error=2, order_reg=1, alpha_reg=0.1) - x_gt))}")
    #
    # with asection("Running l1l1 solver on small matrix: "):
    #     aprint(f"error = {numpy.mean(numpy.absolute(linsolve(a, y_obs, order_error=1, order_reg=1, alpha_reg=0.1) - x_gt))}")
    #
    # with asection("Running l0.5l1 solver on small matrix: "):
    #     aprint(f"error = {numpy.mean(numpy.absolute(linsolve(a, y_obs, order_error=0.5, order_reg=1, alpha_reg=0.1) - x_gt))}")


def _run_solver(w=37, h=41, noise=1e-2, display=False, **kwargs):
    a = random.rand(w, h)
    a *= random.rand(w, h) > 0.9
    x_gt = random.rand(h)
    x_gt *= random.rand(h) > 0.5
    y_obs = a @ x_gt + noise * (random.rand(w) - 0.5)
    x = linsolve(a, y_obs, **kwargs)
    mean_abs_error = numpy.mean(numpy.absolute(x - x_gt))

    if display:
        aprint(f"x_gt  : {x_gt} ")
        aprint(f"x     : {x} ")
        aprint(f"error : {numpy.absolute(x - x_gt)} ")

    return mean_abs_error


def _test_linear_solver_large():
    with asection("Running solver on large matrix: "):
        aprint(_run_solver(w=1280, h=1370, order_error=1, order_reg=1, alpha_reg=0.1, display=True, limited=True))

# def test_linear_solver_compare():
#     n = 16
#
#     mean_error = numpy.mean(list(_run_solver(order_error=2, alpha_reg=0) for _ in range(n)))
#     aprint(f"L2 mean_error={mean_error}")
#
#     mean_error = numpy.mean(list(_run_solver(order_error=1, alpha_reg=0) for _ in range(n)))
#     aprint(f"L1 mean_error={mean_error}")
#
#     mean_error = numpy.mean(list(_run_solver(order_error=0.5, alpha_reg=0) for _ in range(n)))
#     aprint(f"L0.5 mean_error={mean_error}")
#
#     mean_error = numpy.mean(list(_run_solver(order_error=2, order_reg=2, alpha_reg=0.1) for _ in range(n)))
#     aprint(f"L2L2 mean_error={mean_error}")
#
#     mean_error = numpy.mean(list(_run_solver(order_error=2, order_reg=1, alpha_reg=0.1) for _ in range(n)))
#     aprint(f"L2L1 mean_error={mean_error}")
#
#     mean_error = numpy.mean(list(_run_solver(order_error=1, order_reg=1, alpha_reg=0.1) for _ in range(n)))
#     aprint(f"L1L1 mean_error={mean_error}")
#
#     mean_error = numpy.mean(list(_run_solver(order_error=1, order_reg=0.5, alpha_reg=0.1) for _ in range(n)))
#     aprint(f"L1L0.5 mean_error={mean_error}")



import click
from arbol.arbol import aprint, asection

from dexp.cli.dexp_main import _default_clevel
from dexp.cli.dexp_main import _default_workers_backend
from dexp.cli.parsing import _parse_channels, _get_output_path, _parse_slicing
from dexp.datasets.open_dataset import glob_datasets
from dexp.datasets.operations.tiff import dataset_tiff


@click.command()
@click.argument('input_paths', nargs=-1)  # ,  help='input path'
@click.option('--output_path', '-o')  # , help='output path'
@click.option('--channels', '-c', default=None, help='selected channels.')  #
@click.option('--slicing', '-s', default=None, help='dataset slice (TZYX), e.g. [0:5] (first five stacks) [:,0:100] (cropping in z) ')  #
@click.option('--overwrite', '-w', is_flag=True, help='to force overwrite of target', show_default=True)  # , help='dataset slice'
@click.option('--project', '-p', type=int, default=None, help='max projection over given axis (0->T, 1->Z, 2->Y, 3->X)')  # , help='dataset slice'
@click.option('--split', is_flag=True, help='Splits dataset along first dimension, be carefull, if you slice to a single time point this will split along z!')  # , help='dataset slice'
@click.option('--clevel', '-l', type=int, default=_default_clevel, help='Compression level, 0 means no compression, max is 9', show_default=True)  # , help='dataset slice'
@click.option('--workers', '-k', default=-4, help='Number of worker threads to spawn. Negative numbers n correspond to: number_of _cores / |n|', show_default=True)  #
@click.option('--workersbackend', '-wkb', type=str, default=_default_workers_backend, help='What backend to spawn workers with, can be ‘loky’ (multi-process) or ‘threading’ (multi-thread) ', show_default=True)
def tiff(input_paths,
         output_path,
         channels,
         slicing,
         overwrite,
         project,
         split,
         clevel,
         workers,
         workersbackend):
    """ Exports dataset as TIFF file(s).
    """

    input_dataset, input_paths = glob_datasets(input_paths)
    output_path = _get_output_path(input_paths[0], output_path)
    slicing = _parse_slicing(slicing)
    channels = _parse_channels(input_dataset, channels)

    with asection(f"Exporting to TIFF datset: {input_paths}, channels: {channels}, slice: {slicing}, project:{project}, split:{split}"):
        dataset_tiff(input_dataset,
                     output_path,
                     channels=channels,
                     slicing=slicing,
                     overwrite=overwrite,
                     project=project,
                     one_file_per_first_dim=split,
                     clevel=clevel,
                     workers=workers,
                     workersbackend=workersbackend
                     )

        input_dataset.close()
        aprint("Done!")



from skimage.registration import phase_cross_correlation

from dexp.processing.backends.backend import Backend
from dexp.processing.registration.model.translation_registration_model import TranslationRegistrationModel
from dexp.processing.registration.translation_nd import register_translation_nd


def register_translation_2d_skimage(image_a, image_b,
                                    upsample_factor: int = 16,
                                    force_numpy: bool = False,
                                    internal_dtype=None,
                                    **kwargs) -> TranslationRegistrationModel:
    """
    Registers two 2D images using just a translation-only model using skimage registration code.
    Warning: Works well unless images have lots of noise or overlap is partial.

    Parameters
    ----------
    image_a : First image
    image_b : Second Image
    upsample_factor : Upsampling factor for sub-pixel accuracy
    force_numpy : Forces output model to be allocated with numpy arrays.
    internal_dtype : internal dtype for computation
    kwargs : additional optional parameters

    Returns
    -------
    Translation-only registration model

    """
    image_a = Backend.to_numpy(image_a, dtype=internal_dtype)
    image_b = Backend.to_numpy(image_b, dtype=internal_dtype)
    shifts, error, _ = phase_cross_correlation(image_a, image_b, upsample_factor=upsample_factor, **kwargs)
    return TranslationRegistrationModel(shift_vector=shifts, confidence=error, force_numpy=force_numpy)


def register_translation_2d_dexp(image_a, image_b,
                                 **kwargs) -> TranslationRegistrationModel:
    """
    Registers two 2D images using just a translation-only model using dexp own registration code.

    Parameters
    ----------
    image_a : First image
    image_b : Second Image
    kwargs : additional optional parameters

    Returns
    -------
    Translation-only registration model

    """
    return register_translation_nd(image_a, image_b,
                                   **kwargs)



import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.numpy_backend import NumpyBackend


def fft_convolve(image1, image2,
                 mode: str = 'reflect',
                 in_place: bool = True,
                 internal_dtype=None):
    """
    Fast FFT based convolution.

    Parameters
    ----------
    image1 : First image
    image2 : Second image
    mode : Not supported!
    in_place : If true then the two input images might be modified and reused for the result.

    Returns
    -------
    Convolved image: image1 ○ image2
    """
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    if image1.ndim == image2.ndim == 0:  # scalar inputs
        return image1 * image2
    elif not image1.ndim == image2.ndim:
        raise ValueError("Dimensions do not match.")
    elif image1.size == 0 or image2.size == 0:  # empty arrays
        return xp.asarray([])
    elif image1.dtype != image2.dtype:
        raise ValueError("Two images must have same dtype!")

    if internal_dtype is None:
        internal_dtype = image1.dtype

    if type(Backend.current()) is NumpyBackend:
        internal_dtype = numpy.float32

    original_dtype = image1.dtype
    image1 = Backend.to_backend(image1, dtype=internal_dtype, force_copy=False)
    image2 = Backend.to_backend(image2, dtype=internal_dtype, force_copy=False)

    if mode != 'wrap':
        pad_width = tuple((tuple((s // 2, s // 2)) for s in image2.shape))
        image1 = xp.pad(image1, pad_width=pad_width, mode=mode)

    s1 = numpy.asarray(image1.shape)
    s2 = numpy.asarray(image2.shape)

    shape = tuple(s1 + s2 - 1)

    fsize = shape  # tuple(int(2 ** math.ceil(math.log2(x))) for x in tuple(shape))

    image1_fft = sp.fft.rfftn(image1, fsize, overwrite_x=in_place)
    image2_fft = sp.fft.rfftn(image2, fsize, overwrite_x=in_place)
    image1_fft *= image2_fft
    del image2_fft
    result = sp.fft.irfftn(image1_fft, overwrite_x=in_place)
    if not in_place:
        del image1_fft

    fslice = tuple([slice(0, int(sz)) for sz in shape])
    result = result[fslice]

    newshape = numpy.asarray(image1.shape)
    currshape = numpy.array(result.shape)
    startind = (currshape - newshape) // 2
    endind = startind + newshape
    myslice = [slice(startind[k], endind[k]) for k in range(len(endind))]

    result = result[tuple(myslice)]

    if mode != 'wrap':
        slicing = tuple(slice(s // 2, -(s // 2)) for s in image2.shape)
        result = result[slicing]

    result = result.astype(dtype=original_dtype, copy=False)

    return result



import numpy
from skimage.data import camera
from skimage.util import random_noise

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.filters.butterworth_filter import butterworth_filter


def demo_butterworth_filter_numpy():
    with NumpyBackend():
        _demo_butterworth_filter()


def demo_butterworth_filter_cupy():
    try:
        with CupyBackend():
            _demo_butterworth_filter()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _demo_butterworth_filter():
    image = camera().astype(numpy.float32) / 255
    noisy = random_noise(image.copy(), mode="gaussian", var=0.005, seed=0, clip=False)
    # noisy = random_noise(noisy, mode="s&p", amount=0.03, seed=0, clip=False)

    shape = (13, 13)
    cutoffs = 0.75
    n = 8

    filtered = butterworth_filter(noisy,
                                  shape=shape,
                                  cutoffs=cutoffs,
                                  cutoffs_in_freq_units=False,
                                  order=n)

    from napari import Viewer, gui_qt
    with gui_qt():
        def _c(array):
            return Backend.to_numpy(array)

        viewer = Viewer()
        viewer.add_image(_c(image), name='image')
        viewer.add_image(_c(noisy), name='noisy')
        viewer.add_image(_c(filtered), name='filtered')


if __name__ == "__main__":
    demo_butterworth_filter_cupy()
    demo_butterworth_filter_numpy()



from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.synthetic_datasets.multiview_data import generate_fusion_test_data
from dexp.utils.timeit import timeit


def test_multiview_data_numpy():
    with NumpyBackend():
        _test_multiview_data()


def test_multiview_data_cupy():
    try:
        with CupyBackend():
            _test_multiview_data()
    except (ModuleNotFoundError, NotImplementedError):
        print("Cupy module not found! ignored!")


def _test_multiview_data(length_xy=128):
    with timeit("generate data"):
        image_gt, image_lowq, blend_a, blend_b, image1, image2 = generate_fusion_test_data(add_noise=False, length_xy=length_xy, length_z_factor=2)

    assert image_gt.shape == image_lowq.shape
    assert image_gt.shape == image1.shape
    assert image_gt.shape == image2.shape

    # from napari import Viewer, gui_qt
    # with gui_qt():
    #     def _c(array):
    #         return backend.to_numpy(array)
    #     viewer = Viewer()
    #     viewer.add_image(_c(image_gt), name='image_gt')
    #     viewer.add_image(_c(image_lowq), name='image_lowq')
    #     viewer.add_image(_c(blend_a), name='blend_a')
    #     viewer.add_image(_c(blend_b), name='blend_b')
    #     viewer.add_image(_c(image1), name='image1')
    #     viewer.add_image(_c(image2), name='image2')


test_multiview_data_cupy()
test_multiview_data_numpy()



from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.color.demo.demo_time_stamp import demo_time_stamp


def test_time_stamp_numpy():
    with NumpyBackend():
        _test_time_stamp()


def test_time_stamp_cupy():
    try:
        with CupyBackend():
            _test_time_stamp()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _test_time_stamp():
    demo_time_stamp(display=False)



import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.utils.fit_shape import fit_to_shape
from dexp.processing.utils.nan_to_zero import nan_to_zero


def dehaze(image,
           size: int = 21,
           downscale: int = 4,
           minimal_zero_level: float = 0,
           correct_max_level: bool = True,
           in_place: bool = True,
           internal_dtype=None
           ):
    """
    Dehazes an image by means of a non-linear low-pass rejection filter.

    Parameters
    ----------
    image : image to filter
    size : filter size
    downscale : downscale factor for speeding up computation of the haze map.
    minimal_zero_level : minimal zero level to substract
    in_place : True if the input image may be modified in-place.
    internal_dtype : internal dtype for computation

    Returns
    -------
    Dehazed image

    """
    sp = Backend.get_sp_module()
    xp = Backend.get_xp_module()

    if internal_dtype is None:
        internal_dtype = image.dtype

    if type(Backend.current()) is NumpyBackend:
        internal_dtype = numpy.float32

    original_dtype = image.dtype
    image = Backend.to_backend(image, dtype=internal_dtype, force_copy=not in_place)
    # original_image = image.copy()

    minimal_zero_level = xp.asarray(minimal_zero_level, dtype=internal_dtype)

    # get rid of low values due to noise:
    image_zero_level = sp.ndimage.filters.maximum_filter(image, size=3)

    # downscale to speed up the rest of the computation:
    downscaled_image = sp.ndimage.interpolation.zoom(image_zero_level, zoom=1 / downscale, order=0)

    # find min values:
    image_zero_level = sp.ndimage.filters.minimum_filter(downscaled_image, size=max(1, size // downscale))

    # expand reach of these min values:
    image_zero_level = sp.ndimage.filters.maximum_filter(image_zero_level, size=max(1, size // downscale))

    # smooth out:
    image_zero_level = sp.ndimage.filters.gaussian_filter(image_zero_level, sigma=max(1, size // (2 * downscale)))

    # scale up again:
    image_zero_level = sp.ndimage.zoom(image_zero_level, zoom=downscale, order=1)

    # Padding to recover original image size:
    image_zero_level = fit_to_shape(image_zero_level, shape=image.shape)

    # Ensure that we remove at least the minimum zero level:
    if minimal_zero_level > 0:
        image_zero_level = xp.maximum(image_zero_level, minimal_zero_level)

    # remove zero level:
    image -= image_zero_level
    del image_zero_level

    # clip:
    image = xp.maximum(image, 0, out=image)

    if correct_max_level:
        # get image max level before:
        # twice filtering is to match the extent reached for the zero_level image (see above combination of min then max filters)
        downscaled_image = sp.ndimage.filters.maximum_filter(downscaled_image, size=max(1, size // downscale))
        image_max_level_before = sp.ndimage.filters.gaussian_filter(downscaled_image, sigma=max(1, size // downscale))

        # get image max level after:
        downscaled_image_after = sp.ndimage.filters.maximum_filter(image, size=3)
        downscaled_image_after = sp.ndimage.interpolation.zoom(downscaled_image_after, zoom=1 / downscale, order=0)
        image_max_level_after = sp.ndimage.filters.maximum_filter(downscaled_image_after, size=max(1, size // downscale))
        image_max_level_after = sp.ndimage.filters.gaussian_filter(image_max_level_after, sigma=max(1, size // downscale))

        # Correction ratio:
        epsilon = xp.asarray(1e-6, dtype=internal_dtype)
        correction_ratio = image_max_level_before
        correction_ratio /= (image_max_level_after + epsilon)
        correction_ratio = nan_to_zero(correction_ratio, copy=False)
        del image_max_level_after
        correction_ratio = sp.ndimage.zoom(correction_ratio, zoom=downscale, order=1)
        correction_ratio = fit_to_shape(correction_ratio, shape=image.shape)

        image *= correction_ratio
        del correction_ratio

    # convert back to original dtype
    image = image.astype(dtype=original_dtype, copy=False)

    # from napari import gui_qt, Viewer
    # with gui_qt():
    #     def _c(array):
    #         return backend.to_numpy(array)
    #     viewer = Viewer()
    #     viewer.add_image(_c(image), name='original_image')
    #     viewer.add_image(_c(image_zero_level), name='image_zero_level')
    #     viewer.add_image(_c(image), name='dehazed')

    return image



import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.restoration.dehazing import dehaze
from dexp.processing.synthetic_datasets.nuclei_background_data import generate_nuclei_background_data
from dexp.utils.timeit import timeit


def demo_dehaze_numpy():
    with NumpyBackend():
        demo_dehaze_data()


def demo_dehaze_cupy():
    try:
        with CupyBackend():
            demo_dehaze_data()
    except (ModuleNotFoundError, NotImplementedError):
        print("Cupy module not found! ignored!")


def demo_dehaze_data(length_xy=320):
    xp = Backend.get_xp_module()

    with timeit("generate data"):
        image_gt, background, image = generate_nuclei_background_data(add_noise=True,
                                                                      length_xy=length_xy,
                                                                      length_z_factor=4,
                                                                      independent_haze=True,
                                                                      dtype=numpy.float32)

    with timeit('dehaze'):
        dehazed = dehaze(image, size=25, in_place=False, correct_max_level=True)

    # compute percentage of haze removed:
    background_voxels_image = (1 - image_gt) * image
    background_voxels_dehazed = (1 - image_gt) * dehazed
    total_haze = xp.sum(background_voxels_image)
    total_remaining_haze = xp.sum(background_voxels_dehazed)
    percent_removed = (total_haze - total_remaining_haze) / total_haze
    print(f"percent_removed = {percent_removed}")

    # compute error on non-background voxels
    non_background_voxels_image = image_gt * image
    non_background_voxels_dehazed = image_gt * dehazed
    average_error = xp.mean(xp.absolute(non_background_voxels_image - non_background_voxels_dehazed))
    print(f"average_error = {average_error}")

    from napari import gui_qt, Viewer
    with gui_qt():
        def _c(array):
            return Backend.to_numpy(array)

        viewer = Viewer()
        viewer.add_image(_c(image_gt), name='image_gt')
        viewer.add_image(_c(background), name='background')
        viewer.add_image(_c(image), name='image')
        viewer.add_image(_c(dehazed), name='dehazed')
        viewer.grid.enabled = True


if __name__ == "__main__":
    demo_dehaze_cupy()
    demo_dehaze_numpy()



from typing import Optional, Sequence, Tuple

import numpy
from arbol import aprint
from scipy.optimize import minimize

from dexp.processing.backends.backend import Backend


def linsolve(a, y, x0=None,
             maxiter: int = 1e12,
             maxfun: int = 1e12,
             tolerance: float = 1e-6,
             order_error: float = 1,
             order_reg: float = 1,
             alpha_reg: float = 1e-1,
             l2_init: bool = False,
             bounds: Optional[Sequence[Tuple[float, float]]] = None,
             limited: bool = True,
             verbose: bool = False):
    xp = Backend.get_xp_module()

    a = Backend.to_backend(a)
    y = Backend.to_backend(y)

    if x0 is None:
        if l2_init:
            x0 = linsolve(a, y, x0=x0,
                          maxiter=maxiter,
                          tolerance=tolerance,
                          order_error=2,
                          alpha_reg=0,
                          l2_init=False)
        else:
            x0 = numpy.zeros(a.shape[1])

    beta = (1.0 / y.shape[0]) ** (1.0 / order_error)
    alpha = (1.0 / x0.shape[0]) ** (1.0 / order_reg)

    def fun(x):
        x = Backend.to_backend(x)
        if alpha_reg == 0:
            objective = beta * float(xp.linalg.norm(a @ x - y, ord=order_error))
            # aprint(f"objective={objective}, regterm=N/A ")
        else:
            objective = beta * float(xp.linalg.norm(a @ x - y, ord=order_error))
            regularisation_term = (alpha_reg * alpha) * float(xp.linalg.norm(x, ord=order_reg))
            # aprint(f"objective={objective}, regterm={regularisation_term} ")
            objective += regularisation_term
        return objective

    result = minimize(fun,
                      x0,
                      method='L-BFGS-B' if limited else 'BFGS',
                      tol=tolerance,
                      bounds=bounds if limited else None,
                      options={'disp': verbose,
                               'maxiter': maxiter,
                               'maxfun': maxfun,
                               'gtol': tolerance},
                      )
    if result.nit == 0:
        aprint(f"Warning: optimisation finished after {result.nit} iterations!")

    if not result.success:
        raise RuntimeError(f"Convergence failed: '{result.message}' after {result.nit} iterations and {result.nfev} function evaluations.")

    return result.x



import tempfile
from os.path import join

# arr = np.random.rand(1024, 1024)
# serve(arr)  # creates an in-memory store if not zarr.Array or zarr.Group
from dexp.datasets.operations.serve import dataset_serve
from dexp.datasets.zarr_dataset import ZDataset

with tempfile.TemporaryDirectory() as tmpdir:
    print('created temporary directory', tmpdir)

    zdataset = ZDataset(path=join(tmpdir, 'test.zarr'),
                        mode='w',
                        store='dir')

    array = zdataset.add_channel(name='first',
                                 shape=(10, 100, 100, 100),
                                 chunks=(1, 50, 50, 50),
                                 dtype='f4',
                                 codec='zstd',
                                 clevel=3)

    # we initialise almost everything:
    array[...] = 13

    dataset_serve(zdataset)



import click
from arbol.arbol import aprint, asection

from dexp.cli.dexp_main import _default_store, _default_codec, _default_clevel
from dexp.cli.dexp_main import _default_workers_backend
from dexp.cli.parsing import _parse_channels, _get_output_path, _parse_slicing, _parse_devices
from dexp.datasets.open_dataset import glob_datasets
from dexp.datasets.operations.deconv import dataset_deconv


@click.command()
@click.argument('input_paths', nargs=-1)  # ,  help='input path'
@click.option('--output_path', '-o')  # , help='output path'
@click.option('--channels', '-c', default=None, help='list of channels, all channels when ommited.')
@click.option('--slicing', '-s', default=None, help='dataset slice (TZYX), e.g. [0:5] (first five stacks) [:,0:100] (cropping in z) ')
@click.option('--store', '-st', default=_default_store, help='Zarr store: ‘dir’, ‘ndir’, or ‘zip’', show_default=True)
@click.option('--codec', '-z', default=_default_codec, help='compression codec: ‘zstd’, ‘blosclz’, ‘lz4’, ‘lz4hc’, ‘zlib’ or ‘snappy’ ', show_default=True)
@click.option('--clevel', '-l', type=int, default=_default_clevel, help='Compression level', show_default=True)
@click.option('--overwrite', '-w', is_flag=True, help='to force overwrite of target', show_default=True)
@click.option('--chunksize', '-cs', type=int, default=512, help='Chunk size for tiled computation', show_default=True)
@click.option('--method', '-m', type=str, default='lr', help='Deconvolution method: for now only lr (Lucy Richardson)', show_default=True)
@click.option('--iterations', '-i', type=int, default=None, help='Number of deconvolution iterations. More iterations takes longer, will be sharper, but might also be potentially more noisy depending on method. '
                                                                 'The default number of iterations depends on the other parameters, in particular it depends on the choice of backprojection operator. For ‘wb’ as little as 3 iterations suffice. ',
              show_default=True)
@click.option('--maxcorrection', '-mc', type=int, default=None, help='Max correction in folds per iteration. By default there is no limit',
              show_default=True)
@click.option('--power', '-pw', type=float, default=1.0, help='Correction exponent, default for standard LR is 1, set to >1 for acceleration.', show_default=True)
@click.option('--blindspot', '-bs', type=int, default=0, help='Blindspot based noise reduction. Provide size of kernel to use, must be an odd number: 3(recommended), 5, 7. 0 means no blindspot. ', show_default=True)
@click.option('--backprojection', '-bp', type=str, default='tpsf', help='Back projection operator, can be: ‘tpsf’ (transposed PSF = classic) or ‘wb’ (Wiener-Butterworth =  accelerated) ', show_default=True)
@click.option('--objective', '-obj', type=str, default='nikon16x08na', help='Microscope objective to use for computing psf, can be: nikon16x08na or olympus20x10na', show_default=True)
@click.option('--numape', '-na', type=float, default=None, help='Overrides NA value for objective.', show_default=True)
@click.option('--dxy', '-dxy', type=float, default=0.485, help='Voxel size along x and y in microns', show_default=True)
@click.option('--dz', '-dz', type=float, default=4 * 0.485, help='Voxel size along z in microns', show_default=True)
@click.option('--xysize', '-sxy', type=int, default=31, help='PSF size along xy in voxels', show_default=True)
@click.option('--zsize', '-sz', type=int, default=31, help='PSF size along z in voxels', show_default=True)
@click.option('--showpsf', '-sp', is_flag=True, help='Show point spread function (PSF) with napari', show_default=True)
@click.option('--scaling', '-sc', type=str, default='1,1,1', help='Scales input image along the three axis: sz,sy,sx (numpy order). For example: 2,1,1 upscales along z by a factor 2', show_default=True)  #
@click.option('--workers', '-k', type=int, default=-1, help='Number of worker threads to spawn, if -1 then num workers = num devices', show_default=True)
@click.option('--workersbackend', '-wkb', type=str, default=_default_workers_backend, help='What backend to spawn workers with, can be ‘loky’ (multi-process) or ‘threading’ (multi-thread) ', show_default=True)  #
@click.option('--devices', '-d', type=str, default='0', help='Sets the CUDA devices id, e.g. 0,1,2 or ‘all’', show_default=True)  #
@click.option('--check', '-ck', default=True, help='Checking integrity of written file.', show_default=True)  #
def deconv(input_paths,
           output_path,
           channels,
           slicing,
           store,
           codec,
           clevel,
           overwrite,
           chunksize,
           method,
           iterations,
           maxcorrection,
           power,
           blindspot,
           backprojection,
           objective,
           numape,
           dxy, dz,
           xysize, zsize,
           showpsf,
           scaling,
           workers,
           workersbackend,
           devices,
           check):
    """ Deconvolves all or selected channels of a dataset.
    """

    input_dataset, input_paths = glob_datasets(input_paths)
    output_path = _get_output_path(input_paths[0], output_path, "_deconv")

    slicing = _parse_slicing(slicing)
    channels = _parse_channels(input_dataset, channels)
    devices = _parse_devices(devices)

    if ',' in scaling:
        scaling = tuple(float(v) for v in scaling.split(','))

    with asection(f"Deconvolving dataset: {input_paths}, saving it at: {output_path}, for channels: {channels}, slicing: {slicing} "):
        dataset_deconv(input_dataset,
                       output_path,
                       channels=channels,
                       slicing=slicing,
                       store=store,
                       compression=codec,
                       compression_level=clevel,
                       overwrite=overwrite,
                       chunksize=chunksize,
                       method=method,
                       num_iterations=iterations,
                       max_correction=maxcorrection,
                       power=power,
                       blind_spot=blindspot,
                       back_projection=backprojection,
                       objective=objective,
                       numerical_aperture=numape,
                       dxy=dxy,
                       dz=dz,
                       xy_size=xysize,
                       z_size=zsize,
                       show_psf=showpsf,
                       scaling=scaling,
                       workers=workers,
                       workersbackend=workersbackend,
                       devices=devices,
                       check=check
                       )

        input_dataset.close()
        aprint("Done!")



import numpy
from skimage.data import camera

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.interpolation.warp import warp
from dexp.utils.timeit import timeit


def demo_warp_2d_numpy():
    try:
        with NumpyBackend():
            _demo_warp_2d()
    except NotImplementedError:
        print("Numpy version not yet implemented")


def demo_warp_2d_cupy():
    try:
        with CupyBackend():
            _demo_warp_2d()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _demo_warp_2d(grid_size=8):
    image = camera().astype(numpy.float32) / 255
    image = image[0:477, 0:507]

    magnitude = 15
    vector_field = numpy.random.uniform(low=-magnitude, high=+magnitude, size=(grid_size,) * 2 + (2,))

    with timeit("warp"):
        warped = warp(image, vector_field, vector_field_upsampling=4)

    with timeit("dewarped"):
        dewarped = warp(warped, -vector_field, vector_field_upsampling=4)

    from napari import Viewer, gui_qt
    with gui_qt():
        def _c(array):
            return Backend.to_numpy(array)

        viewer = Viewer()
        viewer.add_image(_c(image), name='image')
        viewer.add_image(_c(vector_field), name='vector_field')
        viewer.add_image(_c(warped), name='warped')
        viewer.add_image(_c(dewarped), name='dewarped')


if __name__ == "__main__":
    demo_warp_2d_cupy()
    # demo_warp_2d_numpy()



from abc import ABC, abstractmethod
from typing import Tuple


class PairwiseRegistrationModel(ABC):

    def __init__(self):
        """ Instanciates a translation-only registration model

        """

    @abstractmethod
    def __str__(self):
        pass

    @abstractmethod
    def padding(self):
        raise NotImplementedError("Not implemented")

    @abstractmethod
    def overall_confidence(self) -> float:
        raise NotImplementedError("Not implemented")

    @abstractmethod
    def to_json(self) -> str:
        pass

    @abstractmethod
    def to_numpy(self):
        pass

    @abstractmethod
    def change_relative_to(self, other) -> float:
        pass

    @abstractmethod
    def apply(self, image, pad: bool = False) -> 'Array':
        """ Applies this pairwise registration model to a single image.


        Parameters
        ----------
        image : two images to register to each other.
        pad : Set to True to add padding, False otherwise

        Returns
        -------
        image for which the registration transform has been applied.

        """
        raise NotImplementedError("Not implemented")

    @abstractmethod
    def apply_pair(self, image_a, image_b, pad: bool = False) -> Tuple['Array', 'Array']:
        """ Applies this pairwise registration model to the two given images.
        Two images will be returned as there might be need to pad both images.


        Parameters
        ----------
        image_a, image_b : two images to register to each other.
        pad : Set to True to add padding, False otherwise

        Returns
        -------
        image_a_reg, image_b_reg registered images. In some cases, and if possible, only the second image will be modified.

        """
        raise NotImplementedError("Not implemented")



from arbol import asection
from skimage.color import gray2rgba
from skimage.data import camera

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.color.scale_bar import insert_scale_bar


def demo_scale_bar_numpy():
    with NumpyBackend():
        demo_scale_bar()


def demo_scale_bar_cupy():
    try:
        with CupyBackend():
            demo_scale_bar()
        return True
    except (ModuleNotFoundError, NotImplementedError):
        print("Cupy module not found! ignored!")
        return False


def demo_scale_bar(display=True):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    with asection("generate data"):
        image = Backend.to_backend(gray2rgba(camera()))

    with asection("Apply scale bar..."):
        image_with_scale_bar_br, _ = insert_scale_bar(image,
                                                      length_in_unit=50,
                                                      pixel_scale=0.406,
                                                      translation='bottom_right',
                                                      color=(1, 1, 1, 1),
                                                      )
    with asection("Apply scale bar..."):
        image_with_scale_bar_tl, _ = insert_scale_bar(image,
                                                      length_in_unit=50,
                                                      pixel_scale=0.406,
                                                      translation='top_left',
                                                      color=(1, 1, 1, 1),
                                                      font_size=24,
                                                      )

    with asection("Apply scale bar..."):
        image_with_scale_bar_tr, _ = insert_scale_bar(image,
                                                      length_in_unit=50,
                                                      pixel_scale=0.406,
                                                      translation='top_right',
                                                      color=(1, 0.9, 1, 1),
                                                      mode='add',
                                                      )
    with asection("Apply scale bar..."):
        image_with_scale_bar_bl, _ = insert_scale_bar(image,
                                                      length_in_unit=50,
                                                      pixel_scale=0.406,
                                                      translation='bottom_left',
                                                      color=(1, 0.9, 1, 1),
                                                      )

    if display:
        from napari import Viewer, gui_qt
        with gui_qt():
            def _c(array):
                return Backend.to_numpy(array)

            viewer = Viewer()
            viewer.add_image(_c(image_with_scale_bar_br), name='image_with_scale_bar_br', rgb=True)
            viewer.add_image(_c(image_with_scale_bar_tl), name='image_with_scale_bar_tl', rgb=True)
            viewer.add_image(_c(image_with_scale_bar_tr), name='image_with_scale_bar_tr', rgb=True)
            viewer.add_image(_c(image_with_scale_bar_bl), name='image_with_scale_bar_bl', rgb=True)
            viewer.grid.enabled = True


if __name__ == "__main__":
    if not demo_scale_bar_cupy():
        demo_scale_bar_numpy()



import click
from arbol.arbol import aprint, asection

from dexp.cli.dexp_main import _default_codec, _default_store, _default_clevel
from dexp.cli.dexp_main import _default_workers_backend
from dexp.cli.parsing import _parse_channels, _get_output_path, _parse_slicing, _parse_devices
from dexp.datasets.open_dataset import glob_datasets
from dexp.datasets.operations.fuse import dataset_fuse


@click.command()
@click.argument('input_paths', nargs=-1)  # ,  help='input path'
@click.option('--output_path', '-o')  # , help='output path'
@click.option('--channels', '-c', default=None, help='list of channels for the view in standard order for the microscope type (C0L0, C0L1, C1L0, C1L1,...)')
@click.option('--slicing', '-s', default=None, help='dataset slice (TZYX), e.g. [0:5] (first five stacks) [:,0:100] (cropping in z) ')  #
@click.option('--store', '-st', default=_default_store, help='Zarr store: ‘dir’, ‘ndir’, or ‘zip’', show_default=True)
@click.option('--codec', '-z', default=_default_codec, help='compression codec: ‘zstd’, ‘blosclz’, ‘lz4’, ‘lz4hc’, ‘zlib’ or ‘snappy’ ')
@click.option('--clevel', '-l', type=int, default=_default_clevel, help='Compression level', show_default=True)
@click.option('--overwrite', '-w', is_flag=True, help='to force overwrite of target', show_default=True)  # , help='dataset slice'
@click.option('--microscope', '-m', type=str, default='simview', help='Microscope objective to use for computing psf, can be: simview or mvsols', show_default=True)
@click.option('--equalise/--no-equalise', '-eq/-neq', default=True, help='Equalise intensity of views before fusion, or not.', show_default=True)
@click.option('--equalisemode', '-eqm', default='first', help='Equalisation modes: compute correction ratios only for first time point: ‘first’ or for all time points: ‘all’.', show_default=True)
@click.option('--zerolevel', '-zl', type=int, default=0, help="‘zero-level’ i.e. the pixel values in the restoration (to be substracted)", show_default=True)  #
@click.option('--cliphigh', '-ch', type=int, default=1024, help='Clips voxel values above the given value, if zero no clipping is done', show_default=True)  #
@click.option('--fusion', '-f', type=str, default='tg', help="Fusion mode, can be: ‘tg’ or ‘dct’.  ", show_default=True)  #
@click.option('--fusion_bias_strength', '-fbs', type=(float, float), default=(0.5, 0.02), help='Fusion bias strength for illumination and detection ‘fbs_i fbs_d’, set to ‘0 0’) if fusing a cropped region', show_default=True)  #
@click.option('--dehaze_size', '-dhs', type=int, default=65, help='Filter size (scale) for dehazing the final regsitered and fused image to reduce effect of scattered and out-of-focus light. Set to zero to deactivate.',
              show_default=True)  #
@click.option('--dark_denoise_threshold', '-ddt', type=int, default=0, help='Threshold for denoises the dark pixels of the image -- helps increase compression ratio. Set to zero to deactivate.', show_default=True)  #
@click.option('--zpadapodise', '-zpa', type=(int, int), default=(8, 96),
              help='Pads and apodises the views along z before fusion: ‘pad apo’, where pad is a padding length, and apo is apodisation length, both in voxels. If pad=apo, no original voxel is modified and only added voxels are apodised.',
              show_default=True)  #
@click.option('--loadreg', '-lr', is_flag=True, help='Turn on to load the registration parameters from a previous run', show_default=True)  #
@click.option('--warpregiter', '-wri', type=int, default=4, help='Number of iterations for warp registration (if applicable).', show_default=True)  #
@click.option('--minconfidence', '-mc', type=float, default=0.3, help='Minimal confidence for registration parameters, if below that level the registration parameters for previous time points is used.', show_default=True)  #
@click.option('--maxchange', '-md', type=float, default=16, help='Maximal change in registration parameters, if above that level the registration parameters for previous time points is used.', show_default=True)  #
@click.option('--regedgefilter', '-ref', is_flag=True, help='Use this flag to apply an edge filter to help registration.', show_default=True)  #
@click.option('--hugedataset', '-hd', is_flag=True, help='Use this flag to indicate that the the dataset is _huge_ and that memory allocation should be optimised at the detriment of processing speed.', show_default=True)  #
@click.option('--workers', '-k', type=int, default=-1, help='Number of worker threads to spawn, if -1 then num workers = num devices', show_default=True)  #
@click.option('--workersbackend', '-wkb', type=str, default=_default_workers_backend, help='What backend to spawn workers with, can be ‘loky’ (multi-process) or ‘threading’ (multi-thread) ', show_default=True)  #
@click.option('--devices', '-d', type=str, default='0', help='Sets the CUDA devices id, e.g. 0,1,2 or ‘all’', show_default=True)  #
@click.option('--check', '-ck', default=True, help='Checking integrity of written file.', show_default=True)  #
def fuse(input_paths,
         output_path,
         channels,
         slicing,
         store,
         codec,
         clevel,
         overwrite,
         microscope,
         equalise,
         equalisemode,
         zerolevel,
         cliphigh,
         fusion,
         fusion_bias_strength,
         dehaze_size,
         dark_denoise_threshold,
         zpadapodise,
         loadreg,
         warpregiter,
         minconfidence,
         maxchange,
         regedgefilter,
         hugedataset,
         workers,
         workersbackend,
         devices,
         check):
    """ Fuses the views of a multi-view light-sheet microscope dataset (available: simview and mvsols)
    """

    input_dataset, input_paths = glob_datasets(input_paths)
    output_path = _get_output_path(input_paths[0], output_path, "_fused")

    slicing = _parse_slicing(slicing)
    channels = _parse_channels(input_dataset, channels)
    devices = _parse_devices(devices)

    with asection(f"Fusing dataset: {input_paths}, saving it at: {output_path}, for channels: {channels}, slicing: {slicing} "):
        aprint(f"Microscope type: {microscope}, fusion type: {fusion}")
        aprint(f"Devices used: {devices}, workers: {workers} ")
        dataset_fuse(input_dataset,
                     output_path,
                     channels=channels,
                     slicing=slicing,
                     store=store,
                     compression=codec,
                     compression_level=clevel,
                     overwrite=overwrite,
                     microscope=microscope,
                     equalise=equalise,
                     equalise_mode=equalisemode,
                     zero_level=zerolevel,
                     clip_too_high=cliphigh,
                     fusion=fusion,
                     fusion_bias_strength_i=fusion_bias_strength[0],
                     fusion_bias_strength_d=fusion_bias_strength[1],
                     dehaze_size=dehaze_size,
                     dark_denoise_threshold=dark_denoise_threshold,
                     z_pad_apodise=zpadapodise,
                     loadreg=loadreg,
                     warpreg_num_iterations=warpregiter,
                     min_confidence=minconfidence,
                     max_change=maxchange,
                     registration_edge_filter=regedgefilter,
                     huge_dataset=hugedataset,
                     workers=workers,
                     workersbackend=workersbackend,
                     devices=devices,
                     check=check)

        input_dataset.close()
        aprint("Done!")



import random
import tempfile
from os.path import join

import numpy
from arbol import aprint, asection
from pytest import approx

from dexp.datasets.operations.stabilize import dataset_stabilize
from dexp.datasets.zarr_dataset import ZDataset
from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.interpolation.warp import warp
from dexp.processing.registration.sequence_proj import image_stabilisation_proj
from dexp.processing.synthetic_datasets.nuclei_background_data import generate_nuclei_background_data


def demo_stabilize_numpy():
    with NumpyBackend():
        _demo_stabilize()


def demo_stabilize_cupy():
    try:
        with CupyBackend():
            _demo_stabilize()
    except ModuleNotFoundError:
        aprint("Cupy module not found! demo ignored")


def _demo_stabilize(length_xy=256,
                    n=64,
                    drift_strength=0.8,
                    warp_grid_size=8,
                    warp_strength=2.5,
                    ratio_bad_frames=0.0,
                    additive_noise=0.05,
                    multiplicative_noise=0.1,
                    display=True):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    # used for simulation:
    shifts = []
    vector_fields = []
    x, y, z = 0, 0, 0
    vector_field = xp.zeros((warp_grid_size,) * 3 + (3,), dtype=xp.float32)

    with asection("generate dataset"):
        # prepare simulation data:
        for i in range(n):
            # drift:
            if i < (2 * n // 3):
                x += drift_strength * random.uniform(-1, 1.7)
                y += drift_strength * random.uniform(-1.3, 1)
                z += drift_strength * random.uniform(-1, 1)
            else:
                x += drift_strength * random.uniform(-1.5, 1)
                y += drift_strength * random.uniform(-1, 1.5)
                z += drift_strength * random.uniform(-1, 1)

            # deformation over time:
            vector_field += warp_strength * xp.random.uniform(low=-1, high=+1, size=(warp_grid_size,) * 3 + (3,))

            # simulate sudden imaging jumps:
            if i == n // 2 - n // 3:
                x += 7
                y += -5
                z += 8
            if i == n // 2 + n // 3:
                x += -7
                y += +9
                z += -11

            # keep for later:
            shifts.append((x, y, z))
            vector_fields.append(xp.copy(vector_field))

        # nuclei:
        _, _, image = generate_nuclei_background_data(add_noise=False,
                                                      length_xy=length_xy // 4,
                                                      length_z_factor=1,
                                                      independent_haze=True,
                                                      sphere=True,
                                                      zoom=2,
                                                      dtype=xp.float32)
        image = Backend.to_backend(image)

        # generate reference 'ground truth' timelapse
        image = xp.stack(image for _ in range(n))

        # generate shifted, distorted, degraded timelapse:
        shifted = xp.stack((sp.ndimage.shift(warp(i, vf, vector_field_upsampling=4), shift=s) for i, s, vf in zip(image, shifts, vector_fields)))
        shifted *= xp.clip(xp.random.normal(loc=1, scale=multiplicative_noise, size=shifted.shape, dtype=xp.float32), 0.1, 10)
        shifted += additive_noise * xp.random.rand(*shifted.shape, dtype=xp.float32)

        # simulate dropped, highly corrupted frames:
        for _ in range(int(shifted.shape[0] * ratio_bad_frames)):
            index = random.randint(0, n - 1)
            shifted[index] = sp.ndimage.shift(shifted[index], shift=(random.uniform(-27, 27), random.uniform(-27, 27), random.uniform(-27, 27)))
            shifted[index] += xp.random.rand(*shifted[index].shape)
            shifted[index] *= random.uniform(0.001, 0.5)
            shifted[index] += random.uniform(0.001, 0.1) * xp.random.rand(*shifted[index].shape)

    with tempfile.TemporaryDirectory() as tmpdir:
        aprint('created temporary directory', tmpdir)

        input_path = join(tmpdir, 'input.zarr')
        output_path = join(tmpdir, 'output.zarr')
        input_dataset = ZDataset(path=input_path,
                                 mode='w',
                                 store='dir')

        zarr_input_array = input_dataset.add_channel(name='channel',
                                                     shape=shifted.shape,
                                                     chunks=(1, 50, 50, 50),
                                                     dtype=shifted.dtype,
                                                     codec='zstd',
                                                     clevel=3)
        input_dataset.write_array(channel='channel',
                                  array=Backend.to_numpy(shifted))

        dataset_stabilize(input_dataset=input_dataset,
                          output_path=output_path,
                          channels=['channel']
                          )

        output_dataset = ZDataset(path=output_path,
                                  mode='r',
                                  store='dir')

        zarr_output_array = numpy.asarray(output_dataset.get_array('channel'))

        # stabilise with lower level API:
        model = image_stabilisation_proj(shifted, axis=0, projection_type='max')
        aprint(f"model: {model}")

        # apply stabilisation:
        stabilised_seq_ll = Backend.to_numpy(model.apply_sequence(shifted, axis=0))

        aprint(f"stabilised_seq_ll.shape={stabilised_seq_ll.shape}")
        aprint(f"zarr_output_array.shape={zarr_output_array.shape}")
        assert approx(stabilised_seq_ll.shape[1], zarr_output_array.shape[1], abs=10)
        assert approx(stabilised_seq_ll.shape[2], zarr_output_array.shape[2], abs=10)
        assert approx(stabilised_seq_ll.shape[3], zarr_output_array.shape[3], abs=10)
        error = numpy.mean(numpy.abs(stabilised_seq_ll[:, 0:64, 0:64, 0:64] - zarr_output_array[:, 0:64, 0:64, 0:64]))
        error_null = numpy.mean(numpy.abs(stabilised_seq_ll[:, 0:64, 0:64, 0:64] - Backend.to_numpy(shifted)[:, 0:64, 0:64, 0:64]))
        aprint(f"error={error} versus error_null={error_null}")
        assert error < 20
        assert error < error_null

        if display:
            from napari import Viewer, gui_qt
            with gui_qt():
                def _c(array):
                    return Backend.to_numpy(array)

                viewer = Viewer(ndisplay=2)
                viewer.add_image(_c(image), name='image', colormap='bop orange', blending='additive', visible=True)
                viewer.add_image(_c(shifted), name='shifted', colormap='bop purple', blending='additive', visible=True)
                viewer.add_image(_c(zarr_input_array), name='zarr_input_array', colormap='bop purple', blending='additive', visible=True)
                viewer.add_image(_c(zarr_output_array), name='zarr_output_array', colormap='bop blue', blending='additive', visible=True)
                viewer.add_image(_c(stabilised_seq_ll), name='stabilised_seq_ll', colormap='bop blue', blending='additive', visible=True)
                viewer.grid.enabled = True


if __name__ == "__main__":
    demo_stabilize_cupy()
    # demo_stabilize_numpy()



from arbol import aprint
from numpy import s_


def _get_output_path(input_path, output_path, postfix=''):
    if output_path is None or not output_path.strip():
        return _strip_path(input_path) + postfix
    elif output_path.startswith('_'):
        return _strip_path(input_path)+postfix+output_path
    else:
        return output_path


def _strip_path(path):
    if path.endswith('/') or path.endswith('\\'):
        path = path[:-1]
    if path.endswith('.zip'):
        path = path[:-4]
    if path.endswith('.nested.zarr'):
        path = path[:-12]
    if path.endswith('.zarr'):
        path = path[:-5]
    return path


def _parse_slicing(slicing: str):
    aprint(f"Requested slicing    : '{slicing if slicing else '--All--'}' ")
    if slicing is not None:
        aprint(f"Slicing: {slicing}")
        dummy = s_[1, 2]
        slicing = eval(f"s_{slicing}")
    return slicing


def _parse_channels(input_dataset, channels):
    available_channels = frozenset(input_dataset.channels())
    aprint(f"Available channel(s) : '{available_channels}'")
    aprint(f"Requested channel(s) : '{channels if channels else '--All--'}' ")
    if channels is None:
        channels = input_dataset.channels()
    else:
        channels = tuple(channel.strip() for channel in channels.split(','))
        channels = [channel for channel in channels if channel in available_channels]

    aprint(f"Selected channel(s)  : '{channels}'")
    return channels


def _parse_devices(devices):
    aprint(f"Requested devices    :  '{'--All--' if 'all' in devices else devices}' ")
    if 'all' in devices:
        from dexp.processing.backends.cupy_backend import CupyBackend
        devices = tuple(range(len(CupyBackend.available_devices())))
    else:
        devices = tuple(int(device.strip()) for device in devices.split(','))

    return devices



from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.synthetic_datasets.nuclei_background_data import generate_nuclei_background_data
from dexp.utils.timeit import timeit


def demo_nuclei_background_data_numpy():
    with NumpyBackend():
        demo_nuclei_background_data()


def demo_nuclei_background_data_cupy():
    try:
        with CupyBackend():
            demo_nuclei_background_data()
    except (ModuleNotFoundError, NotImplementedError):
        print("Cupy module not found! ignored!")


def demo_nuclei_background_data(length_xy=320):
    with timeit("generate data"):
        image_gt, background, image = generate_nuclei_background_data(add_noise=True,
                                                                      length_xy=length_xy,
                                                                      length_z_factor=4)

    from napari import Viewer, gui_qt
    with gui_qt():
        def _c(array):
            return Backend.to_numpy(array)

        viewer = Viewer()
        viewer.add_image(_c(image_gt), name='image_gt')
        viewer.add_image(_c(background), name='restoration')
        viewer.add_image(_c(image), name='image')


if __name__ == "__main__":
    demo_nuclei_background_data_cupy()
    demo_nuclei_background_data_numpy()



from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend

from dexp.processing.synthetic_datasets.nuclei_background_data import generate_nuclei_background_data
from dexp.utils.timeit import timeit


def test_nuclei_background_data_numpy():
    with NumpyBackend():
        _test_nuclei_background_data()


def test_nuclei_background_data_cupy():
    try:
        with CupyBackend():
            _test_nuclei_background_data()
    except (ModuleNotFoundError, NotImplementedError):
        print("Cupy module not found! ignored!")


def _test_nuclei_background_data(length_xy=128):
    with timeit("generate data"):
        image_gt, background, image = generate_nuclei_background_data(add_noise=True,
                                                                      length_xy=length_xy,
                                                                      length_z_factor=4)
    assert image_gt.shape == background.shape
    assert image_gt.shape == image.shape

    # from napari import Viewer, gui_qt
    # with gui_qt():
    #     def _c(array):
    #         return backend.to_numpy(array)
    #     viewer = Viewer()
    #     viewer.add_image(_c(image_gt), name='image_gt')
    #     viewer.add_image(_c(image_lowq), name='image_lowq')
    #     viewer.add_image(_c(blend_a), name='blend_a')
    #     viewer.add_image(_c(blend_b), name='blend_b')
    #     viewer.add_image(_c(image1), name='image1')
    #     viewer.add_image(_c(image2), name='image2')


test_nuclei_background_data_cupy()
test_nuclei_background_data_numpy()



from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.registration.demo.demo_warp_2d import _register_warp_2d


# TODO: implement numpy version of warp.
# def test_register_warp_nD_numpy():
#     backend = NumpyBackend()
#     register_warp_nD(backend)


def test_register_warp_2d_cupy():
    try:
        with CupyBackend():
            register_warp_2d()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def register_warp_2d(warp_grid_size=3, reg_grid_size=6):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    image, warped, unwarped, model = _register_warp_2d(warp_grid_size=warp_grid_size, reg_grid_size=reg_grid_size, display=False)

    error_warped = xp.mean(xp.absolute(image - warped))
    error_unwarped = xp.mean(xp.absolute(image - unwarped))
    print(f"error_warped = {error_warped}")
    print(f"error_unwarped = {error_unwarped}")

    assert error_unwarped < error_warped
    assert error_unwarped < 20



import numpy
from numpy.linalg import norm
from scipy.ndimage import convolve
from skimage.data import camera
from skimage.util import random_noise

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.filters.fft_convolve import fft_convolve


def test_fft_convolve_numpy():
    with NumpyBackend():
        _test_fft_convolve()


def test_fft_convolve_cupy():
    try:
        with CupyBackend():
            _test_fft_convolve()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _test_fft_convolve():
    image = camera().astype(numpy.float32) / 255
    noisy = random_noise(image, mode="gaussian", var=0.005, seed=0, clip=False)
    noisy = random_noise(noisy, mode="s&p", amount=0.03, seed=0, clip=False).astype(numpy.float32)

    psf = numpy.asarray([[1, 1, 1], [1, 0, 1], [1, 1, 1]]).astype(numpy.float32)

    result = fft_convolve(noisy, psf)
    reference_result = convolve(noisy, psf)

    result = Backend.to_numpy(result)
    reference_result = Backend.to_numpy(reference_result)

    error = norm(reference_result - result, ord=1) / image.size
    print(error)
    assert error < 1e-3
    #
    # from napari import Viewer, gui_qt
    # with gui_qt():
    #     def _c(array):
    #         return backend.to_numpy(array)
    #
    #     viewer = Viewer()
    #     viewer.add_image(_c(image), name='image')
    #     viewer.add_image(_c(noisy), name='noisy')
    #     viewer.add_image(_c(psf), name='psf')
    #     viewer.add_image(_c(result), name='result')



import numpy
from skimage.data import camera
from skimage.util import random_noise

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.deconvolution.lr_deconvolution import lucy_richardson_deconvolution
from dexp.processing.filters.fft_convolve import fft_convolve
from dexp.processing.filters.kernels.gaussian import gaussian_kernel_nd


def demo_lr_deconvolution_numpy():
    with NumpyBackend():
        _demo_lr_deconvolution()


def demo_lr_deconvolution_cupy():
    try:
        with CupyBackend():
            _demo_lr_deconvolution()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _demo_lr_deconvolution():
    image = camera().astype(numpy.float32) / 255

    with NumpyBackend():
        psf = gaussian_kernel_nd(size=9, ndim=2, dtype=numpy.float32)
        blurry = fft_convolve(image, psf)
        blurry = blurry - blurry.min()
        blurry = blurry / blurry.max()
        noisy = random_noise(blurry, mode="gaussian", var=0.01, seed=0, clip=False)
        noisy = random_noise(noisy, mode="s&p", amount=0.01, seed=0, clip=False)

    iterations = 50

    deconvolved = lucy_richardson_deconvolution(noisy, psf,
                                                num_iterations=iterations,
                                                padding=16)

    deconvolved_blind_spot = lucy_richardson_deconvolution(noisy, psf,
                                                           num_iterations=iterations,
                                                           padding=16,
                                                           blind_spot=5,
                                                           blind_spot_mode='uniform+median')

    from napari import Viewer, gui_qt
    with gui_qt():
        def _c(array):
            return Backend.to_numpy(array)

        viewer = Viewer()
        viewer.add_image(_c(image), name='image')
        viewer.add_image(_c(blurry), name='blurry')
        viewer.add_image(_c(psf), name='psf')
        viewer.add_image(_c(noisy), name='noisy')
        viewer.add_image(_c(deconvolved), name='deconvolved')
        viewer.add_image(_c(deconvolved_blind_spot), name='deconvolved_blind_spot')


if __name__ == "__main__":
    demo_lr_deconvolution_cupy()
    demo_lr_deconvolution_numpy()



# You need to point to a tiff file with 4 views as first dim,
# as produced for example by: dexp tiff -w -s [128:129] dataset.zarr -o /home/royer/Desktop/test_data/test_data.tiff


from arbol import asection, aprint
from napari import gui_qt, Viewer
from tifffile import imread

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.multiview_lightsheet.fusion.simview import simview_fuse_2C2L

filepath = '/home/royer/Desktop/test_data/embryo_4views.tif'


def demo_simview_fusion_numpy():
    with NumpyBackend():
        simview_fusion()


def demo_simview_fusion_cupy():
    try:
        with CupyBackend():
            simview_fusion()
    except ModuleNotFoundError:
        aprint("Cupy module not found! demo ignored")


def simview_fusion():
    with asection("imread"):
        array = imread(filepath)

    C0L0 = array[0]
    C0L1 = array[1]
    C1L0 = array[2]
    C1L1 = array[3]

    with asection("simview_fuse_2I2D"):
        CxLx, model, ratios = simview_fuse_2C2L(C0L0, C0L1, C1L0, C1L1)
    aprint(f"model = {model}, ratios={ratios}")

    with asection("to_numpy"):
        CxLx = Backend.to_numpy(CxLx)

    with gui_qt():
        xp = Backend.get_xp_module()

        def _c(array):
            return Backend.to_numpy(array)

        viewer = Viewer()
        viewer.add_image(_c(C0L0), name='C0L0', contrast_limits=(0, 1000), scale=(4, 1, 1), blending='additive', visible=False)
        viewer.add_image(_c(C0L1), name='C0L1', contrast_limits=(0, 1000), scale=(4, 1, 1), blending='additive', visible=False)
        viewer.add_image(_c(xp.flip(C1L0, -1)), name='C1L0', contrast_limits=(0, 1000), scale=(4, 1, 1), blending='additive', visible=False)
        viewer.add_image(_c(xp.flip(C1L1, -1)), name='C1L1', contrast_limits=(0, 1000), scale=(4, 1, 1), blending='additive', visible=False)
        viewer.add_image(_c(CxLx), name='CxLx', contrast_limits=(0, 1200), scale=(4, 1, 1), blending='additive', colormap='viridis')
        # viewer.add_image(_c(CxLx_deconvolved), name='CxLx_deconvolved', contrast_limits=(0, 1000), scale=(4, 1, 1), blending='additive', colormap='viridis')


if __name__ == "__main__":
    # demo_simview_fuse_numpy()
    demo_simview_fusion_cupy()



import numpy
from scipy.ndimage import convolve
from skimage.data import camera
from skimage.util import random_noise

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.filters.fft_convolve import fft_convolve


def demo_fft_convolve_numpy():
    with NumpyBackend():
        _demo_fft_convolve()


def demo_fft_convolve_cupy():
    try:
        with CupyBackend():
            _demo_fft_convolve()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _demo_fft_convolve():
    image = camera().astype(numpy.float32) / 255
    noisy = random_noise(image, mode="gaussian", var=0.005, seed=0, clip=False)
    noisy = random_noise(noisy, mode="s&p", amount=0.03, seed=0, clip=False).astype(numpy.float32)

    psf = numpy.asarray([[1, 1, 1], [1, 0, 1], [1, 1, 1]]).astype(numpy.float32)

    result = fft_convolve(noisy, psf)
    reference_result = convolve(noisy, psf)

    from napari import Viewer, gui_qt
    with gui_qt():
        def _c(array):
            return Backend.to_numpy(array)

        viewer = Viewer()
        viewer.add_image(_c(image), name='image')
        viewer.add_image(_c(noisy), name='noisy')
        viewer.add_image(_c(psf), name='psf')
        viewer.add_image(_c(reference_result), name='reference_result')
        viewer.add_image(_c(result), name='result')


if __name__ == "__main__":
    demo_fft_convolve_cupy()
    demo_fft_convolve_numpy()



import itertools

from dexp.processing.backends.backend import Backend
from dexp.processing.utils.projection_generator import projection_generator


def center_of_mass(image,
                   mode: str = 'projection',
                   projection_type: str = 'mean',
                   offset_mode: str = 'min',
                   bounding_box: bool = False,
                   ):
    """
    Computes the center of mass of an image.

    Parameters
    ----------
    image: image to compute center of mass of.
    mode: Can be either 'full' or 'projection'. Full mode might lead to high memory consumption depending on the backend, projection mode gives some extra flexibility and is overall safer.
    projection_type: Projection type to use when in 'projection' type: 'mean', 'min', 'max', 'max-min'
    offset_mode: Choice of offset to remove, can be either: 'min', 'median', 'mean', 'middle', and for example: 'p=10' for removing and clipping the lower 10% percentile. Set to N one for no offset removal.
    remove_offset: removes offset to help remove the influence of the background on the center of mass
    bounding_box: if True, the center of mass of the bouning box of non-zero pixels is returned.


    Returns
    -------
    Center of mass as a vector of integers

    """
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    if mode == 'full':
        if offset_mode is not None:
            image = _remove_offset(image, offset_mode, xp)

        # from napari import Viewer, gui_qt
        # with gui_qt():
        #     def _c(array):
        #         return Backend.to_numpy(array)
        #
        #     viewer = Viewer()
        #     viewer.add_image(_c(image), name='image', colormap='viridis', blending='additive', visible=True)

        com = _center_of_mass(image, bounding_box)

    elif mode == 'projection':
        ndim = image.ndim

        com = xp.zeros((ndim,), dtype=xp.float32)
        count = xp.zeros((ndim,), dtype=xp.float32)

        for u, v, projected_image in projection_generator(image,
                                                          projection_type=projection_type,
                                                          nb_axis=2):
            if offset_mode is not None:
                projected_image = _remove_offset(projected_image, offset_mode, xp)

            # from napari import Viewer, gui_qt
            # with gui_qt():
            #     def _c(array):
            #         return Backend.to_numpy(array)
            #
            #     viewer = Viewer()
            #     viewer.add_image(_c(projected_image), name='projected_image', colormap='viridis', blending='additive', visible=True)

            du, dv = _center_of_mass(projected_image, bounding_box)
            com[u] += du
            com[v] += dv
            count[u] += 1
            count[v] += 1

        com /= count

    com = xp.asarray(com, dtype=xp.float32)
    return com


def _center_of_mass(image, bounding_box):
    xp = Backend.get_xp_module()
    if bounding_box:
        bbox = _bounding_box(image)
        com = tuple(0.5 * bbox[2 * a] + 0.5 * bbox[2 * a + 1] for a in range(len(bbox) // 2))
    else:
        sp = Backend.get_sp_module()
        com = sp.ndimage.center_of_mass(image)

    return com


def _bounding_box(image):
    xp = Backend.get_xp_module(image)
    N = image.ndim
    out = []
    for ax in itertools.combinations(reversed(range(N)), N - 1):
        nonzero = xp.any(image, axis=ax)
        out.extend(xp.where(nonzero)[0][[0, -1]])
    return tuple(out)


def _remove_offset(image, offset_mode, xp):
    if offset_mode == 'min':
        image -= image.min()
    elif offset_mode.startswith('p='):
        percentile = float(offset_mode.split('=')[1].strip())
        image -= xp.percentile(image, percentile).astype(image.dtype, copy=False)
    elif offset_mode == 'median':
        image -= xp.median(image).astype(image.dtype, copy=False)
    elif offset_mode == 'mean':
        image -= xp.mean(image).astype(image.dtype, copy=False)
    elif offset_mode == 'middle':
        minv = image.min()
        maxv = image.max()
        offset = 0.5 * (maxv + minv)
        offset = offset.astype(image.dtype)
        image -= offset
    else:
        raise ValueError(f"Unknown offset mode: {offset_mode}")
    image = xp.clip(image, a_min=0, a_max=None, out=image)

    # from napari import Viewer, gui_qt
    # with gui_qt():
    #     def _c(array):
    #         return Backend.to_numpy(array)
    #     viewer = Viewer()
    #     viewer.add_image(_c(image), name='image')

    return image



from arbol.arbol import aprint
from arbol.arbol import asection
from joblib import Parallel, delayed
from zarr.errors import ContainsArrayError, ContainsGroupError

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.best_backend import BestBackend
from dexp.processing.deskew.yang_deskew import yang_deskew


def dataset_deskew(dataset,
                   output_path,
                   channels,
                   slicing,
                   store,
                   compression,
                   compression_level,
                   flips,
                   overwrite,
                   workers,
                   workersbackend,
                   devices,
                   check,
                   stop_at_exception=True):

    arrays = tuple(dataset.get_array(channel, per_z_slice=False, wrap_with_dask=True) for channel in channels)

    with asection(f"Channels:"):
        for view, channel in zip(arrays, channels):
            aprint(f"Channel: {channel} of shape: {view.shape} and dtype: {view.dtype}")

    if slicing is not None:
        aprint(f"Slicing with: {slicing}")
        arrays = tuple(view[slicing] for view in arrays)

    # We allocate last minute once we know the shape...
    from dexp.datasets.zarr_dataset import ZDataset
    mode = 'w' + ('' if overwrite else '-')
    dest_dataset = ZDataset(output_path, mode, store)

    metadata = dataset.get_metadata()
    angle = metadata['angle']
    channel = metadata['channel']
    dz = metadata['dz']
    res = metadata['res']

    for array, channel, flip in zip(arrays, channels, flips):

        # shape and dtype of views to deskew:
        shape = array[0].shape
        nb_timepoints = array.shape[0]

        def process(channel, tp, device):
            try:

                with asection(f"Loading channel {channel} for time point {tp}"):
                    array_tp = array[tp].compute()

                with BestBackend(device, exclusive=True, enable_unified_memory=True):

                    if 'yang' in mode:
                        deskewed_view_tp, = yang_deskew(image=array_tp,
                                                       flip=flip,
                                                       angle=angle,
                                                       dx=res,
                                                       dz=dz)
                    elif 'classic' in mode:
                        raise NotImplementedError('Classic deskewing not yet implemented')
                    else:
                        raise ValueError(f"Deskew mode: {mode} not supported.")

                    with asection(f"Moving array from backend to numpy."):
                        deskewed_view_tp = Backend.to_numpy(deskewed_view_tp, dtype=array.dtype, force_copy=False)

                if channel not in dest_dataset.channels():
                    try:
                        dest_dataset.add_channel(channel,
                                                 shape=array.shape,
                                                 dtype=array.dtype,
                                                 codec=compression,
                                                 clevel=compression_level)
                    except (ContainsArrayError, ContainsGroupError):
                        aprint(f"Other thread/process created channel before... ")

                with asection(f"Saving fused stack for time point {tp}, shape:{deskewed_view_tp.shape}, dtype:{deskewed_view_tp.dtype}"):
                    dest_dataset.write_stack(channel='fused',
                                             time_point=tp,
                                             stack_array=deskewed_view_tp)

                aprint(f"Done processing time point: {tp} .")

            except Exception as error:
                aprint(error)
                aprint(f"Error occurred while processing time point {tp} !")
                import traceback
                traceback.print_exc()

                if stop_at_exception:
                    raise error

        if workers == -1:
            workers = len(devices)
        aprint(f"Number of workers: {workers}")

        if workers > 1:
            Parallel(n_jobs=workers, backend=workersbackend)(delayed(process)(channel, tp, devices[tp % len(devices)]) for tp in range(0, shape[0]))
        else:
            for tp in range(0, nb_timepoints):
                process(channel, tp, devices[0])


    aprint(dest_dataset.info())
    if check:
        dest_dataset.check_integrity()

    # close destination dataset:
    dest_dataset.close()



import numpy as np


class Quaternion(object):
    def __init__(self, w=1., x=0., y=0., z=0.):

        self.data = np.array([w, x, y, z])

    @classmethod
    def copy(cls, rhs):
        return Quaternion(*rhs.data)

    def __getitem__(self, i):
        return self.data[i]

    def __setitem__(self, i, val):
        self.data[i] = val

    def conj(self):
        return Quaternion(self[0], -self[1], -self[2], -self[2])

    def norm(self):
        return np.linalg.norm(self.data)

    def __add__(self, q):
        return Quaternion(*(self.data + q.data))

    def __sub__(self, q):
        return Quaternion(*(self.data - q.data))

    def __mul__(self, q):
        if isinstance(q, Quaternion):
            a1, b1, c1, d1 = self.data
            a2, b2, c2, d2 = q.data
            return Quaternion(a1 * a2 - b1 * b2 - c1 * c2 - d1 * d2,
                              a1 * b2 + b1 * a2 + c1 * d2 - d1 * c2,
                              a1 * c2 - b1 * d2 + c1 * a2 + d1 * b2,
                              a1 * d2 + b1 * c2 - c1 * b2 + d1 * a2)
        else:
            return Quaternion(*(q * self.data))

    def __repr__(self):
        return "Quaternion(%s,%s,%s,%s)" % tuple(self.data)

    def dot(self, q):
        return np.inner(self.data, q.data)

    def normalize(self):
        return Quaternion(*(self.data * 1. / self.norm()))

    def toRotation4(self):
        a, b, c, d = self.data
        return np.array([
            [a ** 2 + b ** 2 - c ** 2 - d ** 2, 2 * (b * c - a * d), 2 * (b * d + a * c), 0],
            [2 * (b * c + a * d), a ** 2 - b ** 2 + c ** 2 - d ** 2, 2 * (c * d - a * b), 0],
            [2 * (b * d - a * c), 2 * (c * d + a * b), a ** 2 - b ** 2 - c ** 2 + d ** 2, 0],
            [0, 0, 0, 1]
        ])

    def toRotation3(self):
        a, b, c, d = self.data
        return np.array([
            [a ** 2 + b ** 2 - c ** 2 - d ** 2, 2 * (b * c - a * d), 2 * (b * d + a * c)],
            [2 * (b * c + a * d), a ** 2 - b ** 2 + c ** 2 - d ** 2, 2 * (c * d - a * b)],
            [2 * (b * d - a * c), 2 * (c * d + a * b), a ** 2 - b ** 2 - c ** 2 + d ** 2],
        ])


def quaternion_slerp(q1, q2, t):
    q1, q2 = q1.normalize(), q2.normalize()

    prod = q1.dot(q2)
    if abs(prod) > .9998:
        return q1 + (q2 - q1) * t
    # picks the shorter great circle
    if prod < 0:
        q2 = q2 * (-1.)
        prod *= -1.

    w = np.arccos(prod)
    return (q1 * (np.sin((1. - t) * w) / np.sin(w))) + q2 * (np.sin(t * w) / np.sin(w))


if __name__ == '__main__':
    q1 = Quaternion(1, 0, 0, 0)
    q2 = Quaternion(0, 1, 0, 0)

    for t in np.linspace(0, 1, 10):
        q = quaternion_slerp(q1, q2, t)
        print(t, q, q.norm())



from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend


def nan_to_zero(array, copy=True):
    """
    Replaces every nan in an array to zero. It might, or not, be able to operate in-place.
    To be safe, the returned array should always be used...

    Parameters
    ----------
    array : array to replace NaNs with zeros.
    copy : True/False to suggest whether copy or in-place behaviour should occur.

    Returns
    -------
    Array for which NaNs have been replace by zero.

    """
    backend = Backend.current()

    if type(backend) is NumpyBackend:
        xp = backend.get_xp_module()
        return xp.nan_to_num(array, copy=copy)
    elif type(backend) is CupyBackend:
        import cupy
        return cupy.nan_to_num(array)



import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.utils.scatter_gather_i2i import scatter_gather_i2i
from dexp.utils.timeit import timeit


def test_scatter_gather_i2i_numpy():
    with NumpyBackend():
        _test_scatter_gather_i2i()


def test_scatter_gather_i2i_cupy():
    try:
        with CupyBackend():
            _test_scatter_gather_i2i(length_xy=512, splits=4, filter_size=7)
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _test_scatter_gather_i2i(ndim=3, length_xy=128, splits=4, filter_size=7):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    image = numpy.random.uniform(0, 1, size=(length_xy,) * ndim)

    def f(x):
        return sp.ndimage.filters.uniform_filter(x, size=filter_size)

    try:
        with timeit("f"):
            result_ref = Backend.to_numpy(f(Backend.to_backend(image)))
    except:
        print("Can't run this, not enough GPU memory!")
        result_ref = 0 * image + 17

    with timeit("scatter_gather(f)"):
        result = scatter_gather_i2i(f, image, chunks=(length_xy // splits,) * ndim, margins=filter_size // 2)

    image = Backend.to_numpy(image)
    result_ref = Backend.to_numpy(result_ref)
    result = Backend.to_numpy(result)

    error = numpy.linalg.norm(result.ravel() - result_ref.ravel(), ord=1) / result_ref.size
    print(f"Error = {error}")

    # from napari import Viewer, gui_qt
    # with gui_qt():
    #     viewer = Viewer()
    #     viewer.add_image(image, name='image')
    #     viewer.add_image(result_ref, name='result_ref')
    #     viewer.add_image(result, name='result')

    assert error < 0.0001



import numpy
from scipy.ndimage import gaussian_filter
from skimage.data import binary_blobs
from skimage.util import random_noise

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.restoration.aap_correction import axis_aligned_pattern_correction
from dexp.utils.timeit import timeit


def add_patterned_noise(image, n):
    image = image.copy()
    image *= 1 + 0.1 * (numpy.random.rand(n, n) - 0.5)
    image += 0.1 * numpy.random.rand(n, n)
    # image += 0.1*numpy.random.rand(n)[]
    image = random_noise(image, mode="gaussian", var=0.00001, seed=0)
    image = random_noise(image, mode="s&p", amount=0.000001, seed=0)
    return image


def test_aap_correction_numpy():
    with NumpyBackend():
        _test_aap_correction()


def test_aap_correction_cupy():
    try:
        with CupyBackend():
            _test_aap_correction()
    except (ModuleNotFoundError, NotImplementedError):
        print("Cupy module not found! ignored!")


def _test_aap_correction(length_xy=128):
    xp = Backend.get_xp_module()

    image = binary_blobs(length=length_xy, seed=1, n_dim=3, volume_fraction=0.01)
    image = image.astype(numpy.float32)
    image = gaussian_filter(image, sigma=4)
    noisy = add_patterned_noise(image, length_xy)

    with timeit("generate data"):
        corrected = axis_aligned_pattern_correction(noisy, in_place=False)
        corrected = Backend.to_numpy(corrected)

    assert corrected is not noisy
    assert corrected.shape == noisy.shape
    assert corrected.dtype == noisy.dtype

    # import napari
    # with napari.gui_qt():
    #     viewer = napari.Viewer()
    #     viewer.add_image(image, name='image')
    #     viewer.add_image(noisy, name='noisy')
    #     viewer.add_image(corrected, name='corrected')

    average_error = numpy.mean(numpy.absolute(image - corrected))
    print(f"average_error = {average_error}")
    assert average_error < 0.07



from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.color.demo.demo_projection import demo_projection


def test_projection_numpy():
    with NumpyBackend():
        _test_projection()


def test_projection_cupy():
    try:
        with CupyBackend():
            _test_projection()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _test_projection():
    demo_projection(length_xy=64, display=False)



from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.registration.translation_nd import register_translation_nd


def test_register_translation_nd_numpy():
    with NumpyBackend():
        _register_translation_nd()


def test_register_translation_nd_cupy():
    try:
        with CupyBackend():
            _register_translation_nd()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _register_translation_nd():
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    image1 = xp.random.uniform(0, 1, size=(257, 237))
    image2 = xp.random.uniform(0, 1, size=(257, 237))

    model = register_translation_nd(image1, image2)
    print(model)

    assert model.confidence < 0.3

    image1 *= 0
    image2 *= 0

    model = register_translation_nd(image1, image2)
    print(model)

    assert model.confidence < 0.3



import tempfile
from os.path import join

from skimage.data import binary_blobs
from skimage.filters import gaussian

from dexp.datasets.joined_dataset import JoinedDataset
from dexp.datasets.zarr_dataset import ZDataset
from dexp.processing.backends.numpy_backend import NumpyBackend


def test_joined_dataset():
    with tempfile.TemporaryDirectory() as tmpdir:
        print('created temporary directory', tmpdir)

        size = 12
        chunks = 6

        dataset_list = []

        for i in range(3):
            zdataset = ZDataset(path=join(tmpdir, f'test{i}.zarr'),
                                mode='w',
                                store='dir')
            dataset_list.append(zdataset)

            zdataset.add_channel(name='first',
                                 shape=(10 + 5 * i, size, size, size),
                                 chunks=(1, chunks, chunks, chunks),
                                 dtype='f4',
                                 codec='zstd',
                                 clevel=2)

            zdataset.add_channel(name='second',
                                 shape=(17 + i // 2, size, size, size),
                                 chunks=None,
                                 dtype='f4',
                                 codec='zstd',
                                 clevel=3)

            with NumpyBackend() as backend:
                xp = backend.get_xp_module()

                blobs = binary_blobs(length=size, n_dim=3, blob_size_fraction=0.1).astype('f4')

                for i in range(0, zdataset.nb_timepoints('first')):
                    blobs = gaussian(blobs, sigma=1 + 0.1 * i)
                    zdataset.write_stack('first', i, blobs)

                for i in range(0, zdataset.nb_timepoints('second')):
                    blobs = gaussian(blobs, sigma=1 + 0.5 * i)
                    zdataset.write_stack('second', i, blobs)

            assert zdataset.check_integrity()

        joined_dataset = JoinedDataset(dataset_list)

        for channel in joined_dataset.channels():
            for dataset in dataset_list:
                assert joined_dataset.nb_timepoints(channel) == sum(dataset.nb_timepoints(channel) for dataset in dataset_list)
                assert joined_dataset.shape(channel)[1:] == dataset.shape(channel)[1:]
                assert joined_dataset.dtype(channel) == dataset.dtype(channel)



import json
from typing import Tuple

import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.interpolation.warp import warp
from dexp.processing.registration.model.pairwise_registration_model import PairwiseRegistrationModel


class WarpRegistrationModel(PairwiseRegistrationModel):

    def __init__(self,
                 vector_field,
                 confidence=None,
                 force_numpy: bool = False):
        """ Warp registration model.
        A warp registration model consists in a vector field of shape (w,1), (h,w,2), or (d,h,w,3) and a confidence matrix of shape (w), (h,w), or (d,h,w),
        The vector field represents a warp transform, and the confidence matrix the confidence score -- within [0, 1] -- of each vector in the vector field.

        Parameters
        ----------
        vector_field : vector field for the warp transform
        confidence : confidence matrix
        force_numpy : when creating this object, you have the option of forcing the use of numpy array instead of the current backend arrays.
        This is usefull to avoid memory fragmentation (e.g. on GPU).
        """
        super().__init__()

        xp = Backend.get_xp_module()

        if force_numpy:
            self.vector_field = Backend.to_numpy(0 if vector_field is None else vector_field)
            self.confidence = Backend.to_numpy(0 if confidence is None else confidence)
        else:
            self.vector_field = xp.asarray(0 if vector_field is None else vector_field)
            self.confidence = xp.asarray(0 if confidence is None else confidence)

    def __str__(self):
        return f"WarpRegistrationModel(vector_field_shape={self.vector_field.shape}, confidence_shape={self.confidence.shape})"

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (self.vector_field == other.vector_field).all() and (self.confidence == other.confidence).all()
        else:
            return False

    def to_json(self) -> str:
        return json.dumps({'type': 'warp', 'vector_field': self.vector_field.tolist(), 'confidence': self.confidence.tolist()})

    def to_numpy(self) -> 'WarpRegistrationModel':
        self.vector_field = Backend.to_numpy(self.vector_field)
        self.confidence = Backend.to_numpy(self.confidence)
        return self

    def padding(self):
        return tuple((0, 0) for _ in range(self.confidence.ndim))

    def overall_confidence(self) -> float:
        return float(self.median_confidence())

    def change_relative_to(self, other) -> float:
        ndim = self.vector_field.ndim
        robust_average = numpy.median(numpy.linalg.norm(Backend.to_numpy(other.vector_field) - Backend.to_numpy(self.vector_field), axis=ndim - 1))
        return float(robust_average)

    def clean(self,
              max_shift: float = None,
              confidence_threshold: float = 0.1,
              mode: str = 'mean'):
        """
        Cleans the vector field by in-filling vectors of low confidence with the median of neighbooring higher-confidence vectors.
        Parameters
        ----------
        confidence_threshold : confidence threshold below which a vector is deamed unreliable.
        mode : How to propagate high-confidence values, can be 'mean' or 'median'
        """
        xp = Backend.get_xp_module()
        sp = Backend.get_sp_module()

        # First we take care of excessive shifts above the 'max_shift' limit:
        vector_field = Backend.to_backend(self.vector_field, force_copy=True)
        original_vector_field = Backend.to_backend(self.vector_field, force_copy=True)
        confidence = Backend.to_backend(self.confidence, force_copy=True)

        if max_shift is not None:
            model_vector_field_norm = xp.linalg.norm(vector_field, axis=-1)
            too_much = model_vector_field_norm > max_shift
            confidence[too_much] = 0.00013579  # close to zero but recognisable

        # Second we fill out low confidence entries:
        num_iterations = 2 * numpy.sum(confidence.shape)  # enough iterations to reach a steady-state
        mask = confidence > confidence_threshold
        vector_field[~mask] = 0
        for i in range(num_iterations):
            if mode == 'median':
                vector_field = sp.ndimage.median_filter(vector_field, size=(3,) * (confidence.ndim) + (1,), mode='nearest')
            elif mode == 'mean':
                vector_field = sp.ndimage.uniform_filter(vector_field, size=(3,) * (confidence.ndim) + (1,), mode='nearest')
            else:
                raise ValueError("Unsupported mode")
            # we make sure to keep the high-confidence vectors unchanged:
            vector_field[mask] = original_vector_field[mask]

        self.vector_field = vector_field
        self.confidence = confidence

    def apply(self,
              image,
              pad: bool = False,
              vector_field_upsampling: int = 2,
              vector_field_upsampling_order: int = 1,
              mode: str = 'border',
              internal_dtype=None
              ) -> 'Array':

        image_warped = warp(image=image,
                            vector_field=self.vector_field,
                            vector_field_upsampling=vector_field_upsampling,
                            vector_field_upsampling_order=vector_field_upsampling_order,
                            mode=mode,
                            internal_dtype=internal_dtype)

        return image_warped

    def apply_pair(self,
                   image_a, image_b,
                   **kwargs) -> Tuple['Array', 'Array']:

        return image_a, self.apply(image_b, **kwargs)

    def median_confidence(self):
        """

        Parameters
        ----------

        Returns
        -------

        """
        xp = Backend.get_xp_module(self.confidence)
        return xp.median(self.confidence)

    def mean_confidence(self):
        """

        Parameters
        ----------
        backend

        Returns
        -------

        """
        xp = Backend.get_xp_module(self.confidence)
        return xp.mean(self.confidence)

    def median_shift_magnitude(self, confidence_threshold: float = 0.5):
        """

        Parameters
        ----------
        backend
        confidence_threshold

        Returns
        -------

        """
        xp = Backend.get_xp_module(self.confidence)
        ndim = self.vector_field.ndim
        vector_field = self.vector_field.astype(dtype=numpy.float32)
        norms = xp.linalg.norm(vector_field, axis=ndim - 1)
        confidence = self.confidence
        norms = norms[confidence > confidence_threshold]
        if norms.size == 0:
            return 0
        elif norms.size == 1:
            return float(norms)
        else:
            return xp.median(norms)



# import dask.array as da
# arr = da.from_zarr("http://localhost:8000/first/first")
#
# from napari import Viewer, gui_qt
#
# with gui_qt():
#     viewer = Viewer()
#     viewer.add_image(arr, name='arr', visible=True)
from dexp.datasets.operations.view import dataset_view

dataset_view('http://localhost:8000')



from typing import Optional

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.filters.sobel_filter import sobel_filter
from dexp.processing.utils.blend_images import blend_images
from dexp.processing.utils.element_wise_affine import element_wise_affine
from dexp.processing.utils.fit_shape import fit_to_shape


def fuse_tg_nd(image_a,
               image_b,
               downscale: Optional[int] = 2,
               sharpness: Optional[float] = 24,
               tenengrad_smoothing: Optional[int] = 4,
               blend_map_smoothing: Optional[int] = 10,
               bias_axis: Optional[int] = None,
               bias_exponent: Optional[float] = 3,
               bias_strength: Optional[float] = 2,
               clip: Optional[bool] = True,
               internal_dtype=None,
               _display_blend_map: bool = False):
    """
    Fuses two images by picking regions from one or the other image based on the local image quality
    measured by using the magnitude of the Sobel gradient -- similarly as in the Tenengrad focus metric.
    A smooth blend map is computed that blends the two images based on local image quality. A bias can be
    introduced to favor one side of an axis versus another.


    Parameters
    ----------
    image_a : First image to fuse
    image_b : Second image to fuse
    downscale : How much to downscale the two images in order to compute the blend map.
    A value of 2 is good to achieve both coarse denoising and reduce compute load.
    sharpness : How 'sharp' should be the choice between the two images.
    A large value makes sure that most of the time the voxel values of one or the other image
    are picked with very little mixing even if local image quality between .
    tenengrad_smoothing : How much to smooth the tenengrad map
    blend_map_smoothing : How much to smooth the blend map
    bias_axis : Along which axis should a bias be introduced in the blend map. None for no bias.
    bias_exponent : Exponent for the bias
    bias_strength : Bias strength -- zero means no bias
    clip : clip output to input images min and max values.
    internal_dtype : dtype for internal computation

    _display_blend_map : For debugging purposes, we can display the images to fuse, the blend map and result.

    Returns
    -------

    """
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    if not image_a.shape == image_b.shape:
        raise ValueError("Arrays must have the same shape")

    if not image_a.dtype == image_b.dtype:
        raise ValueError("Arrays must have the same dtype")

    if internal_dtype is None:
        internal_dtype = image_a.dtype

    if type(Backend.current()) is NumpyBackend:
        internal_dtype = xp.float32

    original_dtype = image_a.dtype

    image_a = Backend.to_backend(image_a, dtype=internal_dtype)
    image_b = Backend.to_backend(image_b, dtype=internal_dtype)

    min_a, max_a = xp.min(image_a), xp.max(image_a)
    min_b, max_b = xp.min(image_b), xp.max(image_b)
    min_value = min(min_a, min_b)
    max_value = min(max_a, max_b)
    del min_a, max_a, min_b, max_b

    # downscale to speed up computation and reduce noise
    d_image_a = sp.ndimage.interpolation.zoom(image_a, zoom=1 / downscale, order=0)
    d_image_b = sp.ndimage.interpolation.zoom(image_b, zoom=1 / downscale, order=0)

    # Denoise further:
    d_image_a = sp.ndimage.gaussian_filter(d_image_a, sigma=1.5)
    d_image_b = sp.ndimage.gaussian_filter(d_image_b, sigma=1.5)

    # Compute Tenengrad filter:
    t_image_a = sobel_filter(d_image_a, exponent=1, normalise_input=False, in_place_normalisation=True)
    t_image_b = sobel_filter(d_image_b, exponent=1, normalise_input=False, in_place_normalisation=True)
    del d_image_a, d_image_b

    # Apply maximum filter:
    t_image_a = sp.ndimage.maximum_filter(t_image_a, size=tenengrad_smoothing)
    t_image_b = sp.ndimage.maximum_filter(t_image_b, size=tenengrad_smoothing)

    # Apply smoothing filter:
    t_image_a = sp.ndimage.uniform_filter(t_image_a, size=max(1, tenengrad_smoothing))
    t_image_b = sp.ndimage.uniform_filter(t_image_b, size=max(1, tenengrad_smoothing))

    # Normalise:

    t_min_value = min(xp.min(t_image_a), xp.min(t_image_b))
    t_max_value = max(xp.max(t_image_a), xp.max(t_image_b))
    alpha = (1 / (t_max_value - t_min_value)).astype(internal_dtype)
    beta = (-t_min_value / (t_max_value - t_min_value)).astype(internal_dtype)
    t_image_a = element_wise_affine(t_image_a, alpha, beta, out=t_image_a)
    t_image_b = element_wise_affine(t_image_b, alpha, beta, out=t_image_b)
    del t_min_value, t_max_value

    # Add bias:
    if bias_axis is not None and bias_strength != 0:
        length = t_image_a.shape[bias_axis]
        bias_vector = xp.linspace(-1, 1, num=length)
        bias_vector = xp.sign(bias_vector) * (xp.absolute(bias_vector) ** bias_exponent)
        new_shape = tuple(s if i == bias_axis else 1 for i, s in enumerate(t_image_a.shape))
        bias_vector = xp.reshape(bias_vector, newshape=new_shape)
        t_image_a -= bias_strength * bias_vector
        t_image_b += bias_strength * bias_vector
        del bias_vector

    # compute the absolute difference and sign:
    diff = t_image_a
    diff -= t_image_b
    del t_image_b

    sgn_diff = xp.sign(diff)
    abs_diff = xp.absolute(diff, out=diff)
    abs_diff **= 1 / sharpness
    del diff

    # compute blending map:
    blend_map = abs_diff
    blend_map *= sgn_diff
    blend_map = element_wise_affine(blend_map, 0.5, 0.5, out=blend_map)
    del sgn_diff

    # Upscale blending map back to original size:
    blend_map = sp.ndimage.zoom(blend_map, zoom=downscale, order=1)

    # Padding to recover original image size:
    blend_map = fit_to_shape(blend_map, shape=image_a.shape)

    # Smooth blend map to have less seams:
    blend_map = sp.ndimage.filters.uniform_filter(blend_map, size=blend_map_smoothing)

    # Fuse using blending map:
    image_fused = blend_images(image_a, image_b, blend_map)
    if not _display_blend_map:
        del image_a, image_b, blend_map

    if clip:
        image_fused = xp.clip(image_fused, min_value, max_value, out=image_fused)

    # Adjust type:
    image_fused = image_fused.astype(original_dtype, copy=False)

    if _display_blend_map:
        from napari import Viewer, gui_qt
        with gui_qt():
            def _c(array):
                return Backend.to_numpy(array)

            viewer = Viewer()
            viewer.add_image(_c(image_a), name='image_a', contrast_limits=(0, 600))
            viewer.add_image(_c(image_b), name='image_b', contrast_limits=(0, 600))
            viewer.add_image(_c(blend_map), name='blend_map')
            viewer.add_image(_c(image_fused), name='image_fused', contrast_limits=(0, 600))

    return image_fused



from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.synthetic_datasets.multiview_data import generate_fusion_test_data
from dexp.utils.timeit import timeit


def demo_multiview_data_numpy():
    with NumpyBackend():
        demo_multiview_data()


def demo_multiview_data_cupy():
    try:
        with CupyBackend():
            demo_multiview_data()
    except (ModuleNotFoundError, NotImplementedError):
        print("Cupy module not found! ignored!")


def demo_multiview_data(length_xy=320):
    with timeit("generate data"):
        image_gt, image_lowq, blend_a, blend_b, image1, image2 = generate_fusion_test_data(add_noise=True, length_xy=length_xy, length_z_factor=4)

    from napari import Viewer, gui_qt
    with gui_qt():
        def _c(array):
            return Backend.to_numpy(array)

        viewer = Viewer()
        viewer.add_image(_c(image_gt), name='image_gt')
        viewer.add_image(_c(image_lowq), name='image_lowq')
        viewer.add_image(_c(blend_a), name='blend_a')
        viewer.add_image(_c(blend_b), name='blend_b')
        viewer.add_image(_c(image1), name='image1')
        viewer.add_image(_c(image2), name='image2')


if __name__ == "__main__":
    demo_multiview_data_cupy()
    demo_multiview_data_numpy()



import multiprocessing
import os
import shutil
from multiprocessing.pool import ThreadPool
from os.path import isfile, isdir, exists
from typing import Tuple, Sequence, Any, Union

import dask
import numpy
import zarr
from arbol.arbol import aprint
from numcodecs import blosc
from zarr import open_group, convenience, CopyError, Blosc, Group

from dexp.datasets.base_dataset import BaseDataset
# Configure multithreading for Dask:
from dexp.processing.backends.backend import Backend

_cpu_count = multiprocessing.cpu_count() // 2
# aprint(f"Number of cores on system: {_cpu_count}")
_nb_threads = max(1, _cpu_count)
dask.config.set(scheduler='threads')
dask.config.set(pool=ThreadPool(_nb_threads))

# Configure multithreading for Blosc:
blosc.set_nthreads(_nb_threads)


# aprint(f"Number of threads used by BLOSC: {blosc.get_nthreads()}")


class ZDataset(BaseDataset):
    _default_chunks = (1, 128, 512, 512)

    def __init__(self, path: str, mode: str = 'r', store: str = None):
        """Instanciates a Zarr dataset (and opens it)

        Parameters
        ----------
        path : path to zarr storage (directory or zip).
        mode : Access mode:
            'r' means read only (must exist);
            'r+' means read/write (must exist);
            'a' means read/write (create if doesn't exist);
            'w' means create (overwrite if exists);
            'w-' means create (fail if exists).
        store : type of store, can be 'dir', 'ndir', or 'zip'

        Returns
        -------
        Zarr dataset


        """

        super().__init__(dask_backed=False)

        self._path = path
        self._store = None
        self._root_group = None
        self._arrays = {}
        self._projections = {}

        # Open remote store:
        if 'http' in path:
            aprint(f"Opening a remote store at: {path}")
            from fsspec import get_mapper
            self.store = get_mapper(path)
            self._root_group = zarr.open(self.store, mode=mode)
            self._initialise_existing()
            return

        # Correct path to adhere to convention:
        if 'a' in mode or 'w' in mode:
            if path.endswith('.zarr.zip') or store == 'zip':
                path = path + '.zip' if path.endswith('.zarr') else path
                path = path if path.endswith('.zarr.zip') else path + '.zarr.zip'
            elif path.endswith('.nested.zarr') or path.endswith('.nested.zarr/') or store == 'ndir':
                path = path if path.endswith('.nested.zarr') else path + '.nested.zarr'
            elif path.endswith('.zarr') or path.endswith('.zarr/') or store == 'dir':
                path = path if path.endswith('.zarr') else path + '.zarr'

        # if exists and overwrite then delete!
        if exists(path) and mode == 'w-':
            raise ValueError(f"Storage '{path}' already exists, add option '-w' to force overwrite!")
        elif exists(path) and mode == 'w':
            aprint(f"Deleting '{path}' for overwrite!")
            if isdir(path):
                shutil.rmtree(path, ignore_errors=True)
            elif isfile(path):
                os.remove(path)

        if exists(path):
            aprint(f"Opening existing Zarr storage: '{path}' with read/write mode: '{mode}' and store type: '{store}'")
            if isfile(path) and (path.endswith('.zarr.zip') or store == 'zip'):
                aprint(f"Opening as ZIP store")
                self._store = zarr.storage.ZipStore(path)
            elif isdir(path) and ((path.endswith('.nested.zarr') or path.endswith('.nested.zarr/') or store == 'ndir')):
                aprint(f"Opening as Nested Directory store")
                self._store = zarr.storage.NestedDirectoryStore(path)
            elif isdir(path) and ((path.endswith('.zarr') or path.endswith('.zarr/') or store == 'dir')):
                aprint(f"Opening as Directory store")
                self._store = zarr.storage.DirectoryStore(path)

            aprint(f"Opening with mode: {mode}")
            self._root_group = open_group(self._store, mode=mode)
            self._initialise_existing()
        elif 'a' in mode or 'w' in mode:
            aprint(f"Creating Zarr storage: '{path}' with read/write mode: '{mode}' and store type: '{store}'")
            if store is None:
                store = 'dir'
            try:
                if path.endswith('.zarr.zip') or store == 'zip':
                    aprint(f"Opening as ZIP store")
                    self._store = zarr.storage.ZipStore(path)
                elif path.endswith('.nested.zarr') or path.endswith('.nested.zarr/') or store == 'ndir':
                    aprint(f"Opening as Nested Directory store")
                    self._store = zarr.storage.NestedDirectoryStore(path)
                elif path.endswith('.zarr') or path.endswith('.zarr/') or store == 'dir':
                    aprint(f"Opening as Directory store")
                    self._store = zarr.storage.DirectoryStore(path)
                else:
                    aprint(f'Cannot open {path}, needs to be a zarr directory (directory that ends with `.zarr` or `.nested.zarr` for nested folders), or a zipped zarr file (file that ends with `.zarr.zip`)')

                self._root_group = zarr.convenience.open(self._store, mode=mode)

            except Exception as e:
                raise ValueError(f"Problem: can't create target file/directory, most likely the target dataset already exists or path incorrect: {path}")
        else:
            raise ValueError(f"Invalid read/write mode or invalid path: {path} (check path!)")

    def _initialise_existing(self):
        self._channels = [channel for channel, _ in self._root_group.groups()]

        aprint(f"Exploring Zarr hierarchy...")
        for channel, channel_group in self._root_group.groups():
            aprint(f"Found channel: {channel}")

            channel_items = channel_group.items()

            for item_name, array in channel_items:
                aprint(f"Found array: {item_name}")

                if item_name == channel or item_name == 'fused':
                    # print(f'Opening array at {path}:{channel}/{item_name} ')
                    self._arrays[channel] = array
                    # self._arrays[channel] = from_zarr(path, component=f"{channel}/{item_name}")
                elif (item_name.startswith(channel) or item_name.startswith('fused')) and '_projection_' in item_name:
                    self._projections[item_name] = array

    def _get_group_for_channel(self, channel: str) -> Union[None, Sequence[Group]]:
        groups = [g for c, g in self._root_group.groups() if c == channel]
        if len(groups) == 0:
            return None
        else:
            return groups[0]

    def close(self):
        # We close the store if it exists, i.e. if we have been writing to the dataset
        if self._store is not None:
            try:
                self._store.close()
            except AttributeError:
                pass

    def check_integrity(self, channels: Sequence[str] = None) -> bool:
        aprint(f"Checking integrity of zarr storage, might take some time.")
        if channels is None:
            channels = self.channels()
        for channel in channels:
            aprint(f"Checking integrity of channel '{channel}'...")
            array = self.get_array(channel, wrap_with_dask=False)
            if array.nchunks_initialized < array.nchunks:
                aprint(f"WARNING! not all chunks initialised! (dtype={array.dtype})")
                return False
            else:
                aprint(f"Channel '{channel}' seems ok!")
                return True

    def channels(self) -> Sequence[str]:
        return list(self._arrays.keys())

    def nb_timepoints(self, channel: str) -> int:
        return self.get_array(channel).shape[0]

    def shape(self, channel: str) -> Sequence[int]:
        return self._arrays[channel].shape

    def chunks(self, channel: str) -> Sequence[int]:
        return self._arrays[channel].chunks

    def dtype(self, channel: str):
        return self.get_array(channel).dtype

    def info(self, channel: str = None) -> str:
        info_str = ''
        if channel:
            info_str += f"Channel: '{channel}', nb time points: {self.shape(channel)[0]}, shape: {self.shape(channel)[1:]}"
            info_str += "\n"
            info_str += str(self._arrays[channel].info)
            return info_str
        else:
            info_str += f"Dataset at location: {self._path}"
            info_str = "Zarr tree: \n"
            info_str += str(self._root_group.tree())
            info_str += "\n\n"
            info_str += "Channels: \n"
            for channel in self.channels():
                info_str += "  └──" + self.info(channel) + "\n\n"

        info_str += "\n\n"
        info_str += "\n\n"
        for name, array in self._arrays.items():
            info_str += f"Channel: {name}  \n"
            info_str += str(array.info)
            info_str += '\n'

        return info_str

    def get_metadata(self):
        """get the attributes stored in the zarr folder"""
        attrs = {}
        for name in self._root_group.attrs:
            attrs[name] = self._root_group.attrs[name]
        return attrs

    def get_array(self, channel: str, per_z_slice: bool = False, wrap_with_dask: bool = False):
        array = self._arrays[channel]
        return dask.array.from_array(array, chunks=array.chunks) if wrap_with_dask else array

    def get_stack(self, channel: str, time_point: int, per_z_slice: bool = False, wrap_with_dask: bool = False):
        stack_array = self.get_array(channel, per_z_slice=per_z_slice, wrap_with_dask=wrap_with_dask)[time_point]
        return stack_array

    def get_projection_array(self, channel: str, axis: int, wrap_with_dask: bool = False) -> Any:
        array = self._projections[self._projection_name(channel, axis)]
        return dask.array.from_array(array, chunks=array.chunks) if wrap_with_dask else array

    def _projection_name(self, channel: str, axis: int):
        return f'{channel}_projection_{axis}'

    def write_stack(self, channel: str, time_point: int, stack_array: numpy.ndarray):
        array_in_zarr = self.get_array(channel=channel,
                                       wrap_with_dask=False)
        array_in_zarr[time_point] = stack_array

        for axis in range(stack_array.ndim):
            xp = Backend.get_xp_module()
            projection = xp.max(stack_array, axis=axis)
            projection_in_zarr = self.get_projection_array(channel=channel,
                                                           axis=axis,
                                                           wrap_with_dask=False)
            projection_in_zarr[time_point] = projection

    def write_array(self, channel: str, array: numpy.ndarray):
        array_in_zarr = self.get_array(channel=channel,
                                       wrap_with_dask=False)
        array_in_zarr[...] = array

        for axis in range(array.ndim - 1):
            xp = Backend.get_xp_module()
            projection = xp.max(array, axis=axis + 1)
            projection_in_zarr = self.get_projection_array(channel=channel,
                                                           axis=axis,
                                                           wrap_with_dask=False)
            projection_in_zarr[...] = projection

    def add_channel(self,
                    name: str,
                    shape: Tuple[int, ...],
                    dtype,
                    chunks: Sequence[int] = None,
                    enable_projections: bool = True,
                    codec: str = 'zstd',
                    clevel: int = 3) -> Any:
        """Adds a channel to this dataset

        Parameters
        ----------
        name : name of channel.
        shape : shape of correspodning array.
        dtype : dtype of array.
        chunks: chunks shape.
        codec: Compression codec to be used ('zstd', 'blosclz', 'lz4', 'lz4hc', 'zlib' or 'snappy').
        clevel: An integer between 0 and 9 specifying the compression level.

        Returns
        -------
        zarr array


        """
        # check if channel exists:
        if name in self.channels():
            raise ValueError("Channel already exist!")

        if chunks is None:
            chunks = ZDataset._default_chunks[0: len(shape)]

        aprint(f"chunks={chunks}")

        # Choosing the fill value to the largest value:
        fill_value = self._get_largest_dtype_value(dtype)

        aprint(f"Adding channel: '{name}' of shape: {shape}, chunks:{chunks}, dtype: {dtype}, fill_value: {fill_value}, codec: {codec}, clevel: {clevel} ")
        compressor = Blosc(cname=codec, clevel=clevel, shuffle=Blosc.BITSHUFFLE)
        filters = []

        channel_group = self._root_group.create_group(name)
        array = channel_group.full(name=name,
                                   shape=shape,
                                   dtype=dtype,
                                   chunks=chunks,
                                   filters=filters,
                                   compressor=compressor,
                                   fill_value=fill_value)

        self._arrays[name] = array

        if enable_projections:
            ndim = len(shape) - 1
            for axis in range(ndim):
                proj_name = self._projection_name(name, axis)

                proj_shape = list(shape)
                del proj_shape[1 + axis]
                proj_shape = tuple(proj_shape)

                # chunking along time must be 1 to allow parallelism, but no chunking for each projection (not needed!)
                proj_chunks = (1,) + (None,) * (len(chunks) - 2)

                proj_array = channel_group.full(name=proj_name,
                                                shape=proj_shape,
                                                dtype=dtype,
                                                chunks=proj_chunks,
                                                filters=filters,
                                                compressor=compressor,
                                                fill_value=fill_value)
                self._projections[proj_name] = proj_array

        return array

    def add_channels_to(self,
                        zdataset: Union[str, 'ZDataset'],
                        channels: Sequence[str],
                        rename: Sequence[str],
                        store: str = None,
                        add_projections: bool = True,
                        overwrite: bool = True,
                        ):
        """Adds channels from this zarr dataset into an other possibly existing zarr dataset

        Parameters
        ----------
        path : zarr dataset or path of zarr dataset.
        channels: list or tuple of channels to add
        rename: list or tuple of new names for channels
        store: type of zarr store: 'dir' or 'zip', only usefull if store does not exist yet!
        add_projections: If True the projections are also copied.
        overwrite: overwrite destination (not fully functional for zip stores!)

        """

        if type(zdataset) is str:
            zdataset = ZDataset(zdataset, 'a', store)

        root = zdataset._root_group

        aprint(f"Existing channels: {zdataset.channels()}")

        for channel, new_name in zip(channels, rename):
            try:
                array = self.get_array(channel, per_z_slice=False, wrap_with_dask=False)
                source_group = self._get_group_for_channel(channel)
                source_arrays = source_group.items()

                aprint(f"Creating group for channel {channel} of new name {new_name}.")
                if new_name not in root.group_keys():
                    dest_group = root.create_group(new_name)
                else:
                    dest_group = root[new_name]

                aprint(f"Fast copying channel {channel} renamed to {new_name} of shape {array.shape} and dtype {array.dtype} ")

                for name, array in source_arrays:
                    if name in self.channels():
                        aprint(f"Fast copying array {name} to {new_name}")
                        convenience.copy(source=array,
                                         dest=dest_group,
                                         name=new_name,
                                         if_exists='replace' if overwrite else 'raise')

                        if add_projections:
                            ndim = array.ndim - 1
                            for axis in range(ndim):
                                proj_array = self.get_projection_array(channel=channel,
                                                                       axis=axis,
                                                                       wrap_with_dask=False)
                                convenience.copy(source=proj_array,
                                                 dest=dest_group,
                                                 name=self._projection_name(new_name, axis),
                                                 if_exists='replace' if overwrite else 'raise')

            except (CopyError, NotImplementedError):
                aprint(f"Channel already exists, set option '-w' to force overwriting! ")

        zdataset.close()



import numpy

from dexp.processing.backends.backend import Backend


def yang_deskew(image,
                flip: bool,
                dx: float,
                dz: float,
                angle: float,
                num_split: int = 4):
    """  Resampling

    Parameters
    ----------
    image  : input image (skewed 3D stack)
    flip   : True for view 0 and False for view 1
    dz     : float, scanning step (stage or galvo scanning step, not the same as the distance between the slices)
    dx     : float, pixel size of the camera
    angle  : float, incident angle of the light sheet, angle between the light sheet and the optical axis
    num_split : number of splits to break down the data into pieces (along y, axis=2) to fit into the memory of GPU

    Returns
    -------

    """
    # we don't want to move the image to the backend just now,
    # as it might be a very large image and we can actually defer moving it to the backend as
    # after splitting...
    xp = Backend.get_xp_module(image)

    if flip:
        image = xp.flip(image, axis=0)

    # rotate the data
    image = xp.rot90(image, k=1, axes=(1, 2))

    # deskew and rotate
    image = resampling_vertical_split(image,
                                      dx=dx,
                                      dz=dz,
                                      angle=angle,
                                      num_split=num_split)

    # flip along axis x
    if flip:
        xp = Backend.get_xp_module(image)
        image = xp.flip(image, axis=2)

    return image


def resampling_vertical_split(image,
                              dz: float = 1.0,
                              dx: float = 0.2,
                              angle: float = 45,
                              num_split: int = 4):
    """ Same as resampling_vertical but splits the input image so that computation can fit in GPU memory.

    Parameters
    ----------
    image   : input image (skewed 3D stack)
    dz     : float, scanning step (stage or galvo scanning step, not the same as the distance between the slices)
    dx     : float, pixel size of the camera
    angle  : float, incident angle of the light sheet, angle between the light sheet and the optical axis
    num_split  : number of splits to break down the data into pieces (along y, axis=2) to fit into the memory of GPU

    Returns
    -------
    Resampled image.
    Important note: for scalability reasons, the returned image is always numpy image.

    """

    if num_split == 1:
        output = resampling_vertical(image, dz, dx, angle=angle)
    else:
        xp = Backend.get_xp_module(image)
        data_gpu_splits = xp.array_split(image, num_split, axis=1)
        for k in range(num_split):
            data_resampled = resampling_vertical(data_gpu_splits[k], dz, dx, angle=angle)
            if k == 0:
                output = Backend.to_numpy(data_resampled)
            else:
                output = numpy.concatenate((output, Backend.to_numpy(data_resampled)), axis=1)

    return output


def resampling_vertical(image,
                        dz: float,
                        dx: float,
                        angle: float):
    """ Resampling of the image by interpolation along vertical direction.
    Here we assume the dz is integer multiple of dx * cos(angle * pi / 180),
    one can also pre-interpolate the data within along the z' axis if this is not the case

    Parameters
    ----------
    image  : input image (skewed 3D stack)
    dz     : float, scanning step (stage or galvo scanning step, not the same as the distance between the slices)
    dx     : float, pixel size of the camera
    angle  : float, incident angle of the light sheet, angle between the light sheet and the optical axis
    num_split  : number of splits to break down the data into pieces (along y, axis=2) to fit into the memory of GPU

    Returns
    -------
    Resampled image


    """
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    (nz, ny, nx) = image.shape
    dtype = image.dtype

    zres = dz * numpy.sin(angle * xp.pi / 180)
    xres = dx * numpy.cos(angle * xp.pi / 180)

    resample_factor = dz / xres
    resample_factor_int = int(round(resample_factor))

    nz_new, ny_new, nx_new = len(range(0, nx, resample_factor_int)), ny, nx + nz * resample_factor_int
    data_reassign = xp.zeros((nz_new, ny_new, nx_new), xp.int16)

    for x in range(nx):
        x_start = x
        x_end = nz * resample_factor_int + x
        data_reassign[x // resample_factor_int, :, x_start:x_end:resample_factor_int] = image[:, :, x].T
    del image

    # rescale the data, interpolate along z
    data_rescale = sp.ndimage.zoom(data_reassign, zoom=(resample_factor_int, 1, 1), order=1)
    del data_reassign

    data_interp = xp.zeros((nz_new, ny_new, nx_new), dtype)

    for z in range(nz_new):
        for k in range(resample_factor_int):
            data_interp[z, :, k::resample_factor_int] = \
                data_rescale[z * resample_factor_int - k, :, k::resample_factor_int]
    del data_rescale

    # rescale the data, to have voxel the same along x an y;
    # remove the first z slice which has artifacts due to resampling
    image_final = sp.ndimage.zoom(data_interp[1:], zoom=(1, 1, xres / dx), order=1)

    return image_final



from typing import Any

import numpy
import scipy
from dask.array import Array

from dexp.processing.backends.backend import Backend


class NumpyBackend(Backend):
    """
    NumpyBackend
    """

    def __init__(self, *args, **kwargs):
        """ Instanciates a Numpy-based Image Processing backend

        """
        ## Important: Leave this, this is to make sure that the ndimage package works properly!
        exec("import scipy.ndimage")

    def copy(self, *args, **kwargs):
        return NumpyBackend()

    def __str__(self):
        return "NumpyBackend"

    def __enter__(self):
        return super().__enter__()

    def __exit__(self, type, value, traceback):
        super().__exit__(type, value, traceback)

    def clear_memory_pool(self):
        pass

    def _to_numpy(self, array, dtype=None, force_copy: bool = False) -> numpy.ndarray:

        if isinstance(array, Array):
            return self._to_numpy(array.compute(), dtype=dtype, force_copy=force_copy)

        if dtype:
            return array.astype(dtype, copy=force_copy)
        elif force_copy:
            return numpy.asarray(array.copy())
        else:
            return numpy.asarray(array)

    def _to_backend(self, array, dtype=None, force_copy: bool = False) -> Any:

        if isinstance(array, Array):
            return self._to_backend(array.compute(), dtype=dtype, force_copy=force_copy)

        if dtype:
            return array.astype(dtype, copy=force_copy)
        elif force_copy:
            return array.copy()
        else:
            return array

    def _get_xp_module(self, array=None) -> Any:
        if array is None:
            return numpy
        else:
            try:
                import cupy
                return cupy.get_array_module(array)
            except ModuleNotFoundError:
                return numpy

    def _get_sp_module(self, array=None) -> Any:
        if array is None:
            return scipy
        else:
            try:
                import cupyx
                return cupyx.scipy.get_array_module(array)
            except ModuleNotFoundError:
                return numpy



from typing import Any

import numpy
import torch
from torch.types import Device

from dexp.processing.backends.backend import Backend


class PytorchBackend(Backend):

    def __init__(self, device: Device = 'cpu'):
        """ Instanciates a Pytorch-based Image Processing backend

        """
        self.device = device

    def __str__(self):
        return "PytorchBackend"

    def close(self):
        # Nothing to do
        pass

    def to_numpy(self, array, dtype=None, force_copy: bool = False) -> numpy.ndarray:
        if torch.is_tensor(array):
            array = array.cpu().detach().numpy()
            if dtype:
                return array.astype(dtype, copy=force_copy)
            elif force_copy:
                return array.copy()
            else:
                return array
        else:
            return array.astype(dtype, copy=False)

    def to_backend(self, array, dtype=None, force_copy: bool = False) -> Any:

        array = self.to_numpy(array)

        # TODO: handle better dtype conversion...
        if torch.is_tensor(array):
            if force_copy:
                return array.detach().clone()
            else:
                return array
        else:
            if dtype:
                array = array.astype(dtype, copy=force_copy)
            return torch.tensor(array, requires_grad=False, device=self.device)

    def get_xp_module(self, array=None) -> Any:
        raise NotImplementedError("numpy-like module not defined for Pytorch backend")

    def get_sp_module(self, array=None) -> Any:
        raise NotImplementedError("numpy-like module not defined for Pytorch backend")



from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.utils.nan_to_zero import nan_to_zero


def test_nan_to_zero_numpy():
    with NumpyBackend():
        _test_nan_to_zero_shape()


def test_nan_to_zero_cupy():
    try:
        with CupyBackend():
            _test_nan_to_zero_shape()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _test_nan_to_zero_shape(length_xy=128):
    xp = Backend.get_xp_module()

    array_1 = xp.random.uniform(0, 1, size=(32, 10, 17)).astype(dtype=xp.float32)
    array_2 = xp.random.uniform(0, 1, size=(32, 10, 17)).astype(dtype=xp.float32)
    array_2[array_2 < 0.1] = 0

    array_1 /= array_2

    assert xp.isinf(array_1).any()  # xp.isnan(array_1).any() or
    array_1 = nan_to_zero(array_1)
    assert not xp.isinf(array_1).any()  # not xp.isnan(array_1).any() and



import os
from os.path import join, exists

import yaml

__config = None


def get_config():
    if __config is not None:
        return __config

    dexp_home = join(os.getenv("HOME"), '.dexp')
    os.makedirs(dexp_home, exist_ok=True)

    config_file_path = join(dexp_home, 'config.yaml')
    if exists(config_file_path):
        with open(config_file_path, 'r') as stream:
            config = yaml.load(stream, Loader=yaml.FullLoader)
            return config

    else:
        config = {}
        return config


__config = get_config()


def get_telegram_bot_token():
    return get_config()['dexpbot']['token'].strip()


def get_telegram_bot_password():
    return get_config()['dexpbot']['password'].strip()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: To use the 'upload' functionality of this file, you must:
#   $ pipenv install twine --dev

import atexit
import io
import os
import sys
from os.path import exists, join
from shutil import rmtree

from setuptools import find_packages, setup, Command
from setuptools.command.install import install


# Package meta-data.
NAME = 'dexp'
DESCRIPTION = 'Light-sheet Dataset EXploration and Processing'
URL = 'https://github.com/royerlab/dexp'
EMAIL = 'jordao.bragantini@czbiohub.org, ahmetcan.solak@czbiohub.org, loic.royer@czbiohub.org'
AUTHOR = 'Jordao Bragantini, Ahmet Can Solak, Loic A Royer'
REQUIRES_PYTHON = '>=3.8.0'

from datetime import datetime
now = datetime.now()
seconds_since_midnight = (now - now.replace(hour=0, minute=0, second=0, microsecond=0)).total_seconds()
minutes_since_midnight = int(seconds_since_midnight//60)
ten_minutes_since_midnight = int(seconds_since_midnight//600)
VERSION = datetime.today().strftime('%Y.%m.%d')+f'.{minutes_since_midnight}'
print(f"Version: {VERSION}")

CUPY_VERSION = '9.0.0b3'

# What packages are required for this module to be executed?
REQUIRED = [
    'numpy>=1.20',
    'scipy',
    'numba',
    'numcodecs',
    'scikit-image',
    'tifffile',
    'numexpr',
    'joblib',
    'pytest',
    'graphviz',
    'click',
    'dask',
    'dask-image',
    'distributed',
    'zarr',
    'cachey',
    'gputil',
    'gpustat',
    'arbol',
    'simple-zarr-server',
    'requests',
    'aiohttp',
    'colorcet',
    'python-telegram-bot',
    'PyYAML',
    'cairocffi',
]

# What packages are optional?
EXTRAS = {
    'source': [f'cupy=={CUPY_VERSION}', ],
    'cuda112': [f'cupy-cuda112=={CUPY_VERSION}', ],
    'cuda111': [f'cupy-cuda111=={CUPY_VERSION}', ],
    'cuda110': [f'cupy-cuda110=={CUPY_VERSION}', ],
    'cuda102': [f'cupy-cuda102=={CUPY_VERSION}', ],
    'cuda101': [f'cupy-cuda101=={CUPY_VERSION}', ],
    'cuda100': [f'cupy-cuda100=={CUPY_VERSION}', ],
    'color': ['colorama', 'ansicolors',],
    'napari': ['napari[all]',]
}





# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
    with open(os.path.join(here, project_slug, '__version__.py')) as f:
        exec(f.read(), about)
else:
    about['__version__'] = VERSION


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):

        # install twine if not
        os.system('pip install -y twine --dev')

        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status(f'Uploading version `{about["__version__"]}` of the package to PyPI via Twine…')
        os.system('twine upload dist/*')

        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(about['__version__']))
        os.system('git push --tags')

        sys.exit()



# Where the magic happens:
setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    # If your package is a single module, use this instead of 'packages':
    # py_modules=['mypackage'],

    entry_points={
    'console_scripts': ['dexp=dexp.cli.dexp_main:cli',
                        'video=dexp.cli.video_main:cli',
                        'install=dexp.cli.install_main:cli']
    },
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    license='BSD 3-Clause',
    license_file='LICENSE.txt',
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Environment :: GPU :: NVIDIA CUDA :: 11.1',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Image Processing',
        'Topic :: Scientific/Engineering :: Visualization'
    ],
    # $ setup.py publish support.
    cmdclass={
        'upload': UploadCommand
    },
)



